import numpy as np


class GeneralizedInvariantsModel:
    r"""Generalized-invariants isotropic hyperelastic material formulation based on the
    principal stretches.

    ..  math::

        \psi = \psi \left(
            I_1\left( E_1, E_2, E_3 \right),
            I_2\left( E_1, E_2, E_3 \right),
            I_3\left( E_1, E_2, E_3 \right) \right)

    The three principal invariants

    ..  math::

        J_1 &= E_1 + E_2 + E_3

        J_2 &= E_1 E_2 + E_2 E_3 + E_1 E_3

        J_3 &= E_1 E_2 E_3

    are formulated on a one-dimensional strain-stretch relation.

    ..  math::

        E_\alpha &= f(\lambda_\alpha)

        E'_\alpha &= f'(\lambda_\alpha) = \frac{\partial f(\lambda_\alpha)}
            {\partial \lambda_\alpha}

        E''_\alpha &= f''(\lambda_\alpha) = \frac{\partial^2 f(\lambda_\alpha)}
            {\partial \lambda_\alpha~\partial \lambda_\alpha}

    Depending on the strain-stretch relation, the invariants contain deformation-
    independent values.

    ..  math::

        J_{1,0} &= J_1(E_\alpha(\lambda_\alpha=1))

        J_{2,0} &= J_2(E_\alpha(\lambda_\alpha=1))

        J_{3,0} &= J_3(E_\alpha(\lambda_\alpha=1))

    The deformation-dependent parts of the invariants are scaled by deformation-
    independent coefficients of normalization. The deformation-independent parts are
    re-added after the scaling.

    ..  math::

        I_1 &= c_1 (J_1 - J_{1,0}) + J_{1,0}

        I_2 &= c_2 (J_2 - J_{2,0}) + J_{2,0}

        I_3 &= J_3

    Note that the scaling is only applied to the first and second invariant, as the
    third invariant does not contribute to the strain energy function at the undeformed
    state.

    ..  math::

        E_0 &= E(\lambda=1)

        E'_0 &= E'(\lambda=1)

        E''_0 &= E''(\lambda=1)

    The second partial derivative of the strain w.r.t. the stretch must be
    provided for a reference strain, e.g. the Green-Lagrange strain measure (at the
    undeformed state).

    ..  math::

        J''_{1,0} &= \frac{3}{2} \left( E''_0 + E'_0 \right)

        J''_{2,0} &= \frac{3}{2} \left( (2 E_0 (E''_0 + E'_0)) - E'^2_0 \right)

    ..  math::

        c_1 &= \frac{J''_{1,0,ref}}{J''_{1,0}}

        c_2 &= \frac{J''_{2,0,ref}}{J''_{2,0}}

    The first partial derivatives of the strain energy function w.r.t. the invariants

    ..  math::

        \psi_{,1} &= \frac{\partial \psi}{\partial I_1}

        \psi_{,2} &= \frac{\partial \psi}{\partial I_2}

        \psi_{,3} &= \frac{\partial \psi}{\partial I_3}

    and the partial derivatives of the invariants w.r.t. the principal stretches are
    defined. From here on, this is consistent with any invariant-based hyperelastic
    material formulation, except for the factors of normalization.

    ..  math::

        \frac{\partial I_1}{\partial E_\alpha} &= c_1

        \frac{\partial I_2}{\partial E_\alpha} &= c_2 \left( E_\beta + E_\gamma \right)

        \frac{\partial I_3}{\partial E_\alpha} &= E_\beta E_\gamma

    The first partial derivatives of the strain energy density w.r.t. the
    principal stretches are required for the principal values of the stress.

    ..  math::

        \frac{\partial \psi}{\partial \lambda_\alpha} =
            \frac{\partial \psi}{\partial I_1} \frac{\partial I_1}{\partial E_\alpha}
            \frac{\partial E_\alpha}{\partial \lambda_\alpha}
            + \frac{\partial \psi}{\partial I_2} \frac{\partial I_2}{\partial E_\alpha}
            \frac{\partial E_\alpha}{\partial \lambda_\alpha}
            + \frac{\partial \psi}{\partial I_3} \frac{\partial I_3}{\partial E_\alpha}
            \frac{\partial E_\alpha}{\partial \lambda_\alpha}


    Furthermore, the second partial derivatives of the strain energy density w.r.t. the
    principal stretches, necessary for the principal components of the elastic tangent
    moduli, are carried out. This is done in two steps: first, the second partial
    derivatives w.r.t. the principal strain components are carried out, followed by the
    projection to the derivatives w.r.t. the principal stretches.

    ..  math::

        \frac{\partial^2 \psi}{\partial E_\alpha~\partial E_\beta} &=
            \frac{\partial^2 \psi}{\partial I_1~\partial I_1}
            \frac{\partial I_1}{\partial E_\alpha}
            \frac{\partial I_1}{\partial E_\beta}
            +
            \frac{\partial^2 \psi}{\partial I_2~\partial I_2}
            \frac{\partial I_2}{\partial E_\alpha}
            \frac{\partial I_2}{\partial E_\beta}
            +
            \frac{\partial^2 \psi}{\partial I_3~\partial I_3}
            \frac{\partial I_3}{\partial E_\alpha}
            \frac{\partial I_3}{\partial E_\beta}

            &+
            \frac{\partial^2 \psi}{\partial I_1~\partial I_2}
            \left(
                \frac{\partial I_1}{\partial E_\alpha}
                \frac{\partial I_2}{\partial E_\beta}
                +
                \frac{\partial I_2}{\partial E_\alpha}
                \frac{\partial I_1}{\partial E_\beta}
            \right)

            &+
            \frac{\partial^2 \psi}{\partial I_2~\partial I_3}
            \left(
                \frac{\partial I_2}{\partial E_\alpha}
                \frac{\partial I_3}{\partial E_\beta}
                +
                \frac{\partial I_3}{\partial E_\alpha}
                \frac{\partial I_2}{\partial E_\beta}
            \right)

            &+
            \frac{\partial^2 \psi}{\partial I_1~\partial I_3}
            \left(
                \frac{\partial I_1}{\partial E_\alpha}
                \frac{\partial I_3}{\partial E_\beta}
                +
                \frac{\partial I_3}{\partial E_\alpha}
                \frac{\partial I_1}{\partial E_\beta}
            \right)

            &+
            \frac{\partial \psi}{\partial I_1}
            \frac{\partial^2 I_1}{\partial E_\alpha~\partial E_\beta}
            +
            \frac{\partial \psi}{\partial I_2}
            \frac{\partial^2 I_1}{\partial E_\alpha~\partial E_\beta}
            +
            \frac{\partial \psi}{\partial I_3}
            \frac{\partial^2 I_1}{\partial E_\alpha~\partial E_\beta}

    ..  math::

        \frac{\partial^2 \psi}{\partial \lambda_\alpha~\partial \lambda_\beta} =
            \frac{\partial E_\alpha}{\partial \lambda_\alpha}
            \frac{\partial^2 \psi}{\partial E_\alpha~\partial E_\beta}
            \frac{\partial E_\beta} {\partial \lambda_\beta}
            +
            \left(
            \frac{\partial \psi}{\partial I_1} \frac{\partial I_1}{\partial E_\alpha}
            +
            \frac{\partial \psi}{\partial I_2} \frac{\partial I_2}{\partial E_\alpha}
            +
            \frac{\partial \psi}{\partial I_3} \frac{\partial I_3}{\partial E_\alpha}
            \right)
            \frac{\partial^2 E_\alpha}{\partial \lambda_\alpha \partial \lambda_\alpha}

    """

    def __init__(self, material, fun, **kwargs):
        """Initialize the generalized invariant-based isotropic hyperelastic material
        formulation."""

        self.material = material
        self.strain = fun
        self.kwargs = kwargs

        E, dEdλ, d2Edλdλ, E0, dEdλ0, d2Edλdλ0 = self.strain(1, **self.kwargs)

        # normalize invariants
        c1_upper = 3 / 2 * (d2Edλdλ0 + dEdλ0)
        c1_lower = 3 / 2 * (d2Edλdλ + dEdλ)

        c2_upper = 3 * (d2Edλdλ0 + dEdλ0) * E0 - 3 / 2 * dEdλ0**2
        c2_lower = 3 * (d2Edλdλ + dEdλ) * E - 3 / 2 * dEdλ**2

        self.c1 = c1_upper / c1_lower
        self.c2 = c2_upper / c2_lower

        self.I10 = 3 * E
        self.I20 = 3 * E**2
        self.I30 = E**3

    def gradient(self, stretches, statevars):
        """The gradient as the partial derivative of the strain energy function w.r.t.
        the principal stretches."""

        # principal strains
        self.E, self.dEdλ, self.d2Edλdλ = self.strain(stretches, **self.kwargs)[:3]
        E = self.E

        # strain invariants
        I1 = E[0] + E[1] + E[2]
        I2 = E[0] * E[1] + E[1] * E[2] + E[2] * E[0]
        I3 = E[0] * E[1] * E[2]

        self.I1 = self.c1 * I1 - self.I10 * (self.c1 - 1)
        self.I2 = self.c2 * I2 - self.I20 * (self.c2 - 1)
        self.I3 = I3

        self.dWdI1, self.dWdI2, self.dWdI3, statevars_new = self.material.gradient(
            self.I1, self.I2, self.I3, statevars
        )

        Eβ = E[[1, 0, 0]]
        Eγ = E[[2, 2, 1]]

        self.dI1dE = self.c1 * np.ones_like(E)
        self.dI2dE = self.c2 * (Eβ + Eγ)
        self.dI3dE = Eβ * Eγ

        dWdλ = np.zeros_like(stretches)

        if self.dWdI1 is not None:
            dWdλ += self.dWdI1 * self.dI1dE * self.dEdλ

        if self.dWdI2 is not None:
            dWdλ += self.dWdI2 * self.dI2dE * self.dEdλ

        if self.dWdI3 is not None:
            dWdλ += self.dWdI3 * self.dI3dE * self.dEdλ

        return dWdλ, statevars_new

    def hessian(self, stretches, statevars):
        """The hessian as the second partial derivatives of the strain energy function
        w.r.t. the principal stretches."""

        dWdλα, statevars_new = self.gradient(stretches, statevars)
        (
            d2WdI1dI1,
            d2WdI2dI2,
            d2WdI3dI3,
            d2WdI1dI2,
            d2WdI2dI3,
            d2WdI1dI3,
        ) = self.material.hessian(self.I1, self.I2, self.I3, statevars)

        dI1dE = self.dI1dE
        dI2dE = self.dI2dE
        dI3dE = self.dI3dE
        dEdλ = self.dEdλ
        d2Edλdλ = self.d2Edλdλ

        Eγ = [2, 0, 1]

        d2WdEαdEβ = np.zeros((6, *dWdλα.shape[1:]))

        α = [0, 1, 2, 0, 1, 0]
        β = [0, 1, 2, 1, 2, 2]

        if d2WdI1dI1 is not None:
            d2WdEαdEβ += d2WdI1dI1 * dI1dE[α] * dI1dE[β]

        if d2WdI2dI2 is not None:
            d2WdEαdEβ += d2WdI2dI2 * dI2dE[α] * dI2dE[β]

        if d2WdI3dI3 is not None:
            d2WdEαdEβ += d2WdI3dI3 * dI3dE[α] * dI3dE[β]

        if d2WdI1dI2 is not None:
            d2WdEαdEβ += d2WdI1dI2 * (dI1dE[α] * dI2dE[β] + dI2dE[α] * dI1dE[β])

        if d2WdI2dI3 is not None:
            d2WdEαdEβ += d2WdI2dI3 * (dI2dE[α] * dI3dE[β] + dI3dE[α] * dI2dE[β])

        if d2WdI1dI3 is not None:
            d2WdEαdEβ += d2WdI1dI3 * (dI1dE[α] * dI3dE[β] + dI3dE[α] * dI1dE[β])

        # if self.dWdI1 is not None:
        #     d2I1dEαdEβ = 0
        #     d2WdEαdEβ[3:] += self.dWdI2 * d2I1dEαdEβ

        if self.dWdI2 is not None:
            d2I2dEαdEβ = self.c2
            d2WdEαdEβ[3:] += self.dWdI2 * d2I2dEαdEβ

        if self.dWdI3 is not None:
            d2I3dEαdEβ = Eγ
            d2WdEαdEβ[3:] += self.dWdI3 * d2I3dEαdEβ

        # in-place modification (avoid the creation of a new array)
        d2Wdλαdλβ = d2WdEαdEβ
        d2Wdλαdλβ *= dEdλ[α] * dEdλ[β]

        if self.dWdI1 is not None:
            d2Wdλαdλβ[:3] += self.dWdI1 * dI1dE * d2Edλdλ

        if self.dWdI2 is not None:
            d2Wdλαdλβ[:3] += self.dWdI2 * dI2dE * d2Edλdλ

        if self.dWdI3 is not None:
            d2Wdλαdλβ[:3] += self.dWdI3 * dI3dE * d2Edλdλ

        return d2Wdλαdλβ



from ._voigt import (
    astensor,
    asvoigt,
    cdya,
    cdya_ik,
    cdya_il,
    ddot,
    det,
    dev,
    dot,
    dya,
    eigh,
    eye,
    inv,
    trace,
    transpose,
    tril_from_triu,
    triu_from_tril,
)

__all__ = [
    "astensor",
    "asvoigt",
    "cdya",
    "cdya_ik",
    "cdya_il",
    "ddot",
    "det",
    "dev",
    "dot",
    "dya",
    "eye",
    "inv",
    "trace",
    "transpose",
    "eigh",
    "tril_from_triu",
    "triu_from_tril",
]



import matplotlib.pyplot as plt


class LabPlotter:
    "Class with methods for generating plots of experiments and simulations."

    def __init__(self):
        self.label = None

    def plot_force_displacement(
        self,
        *args,
        xlabel=r"Displacement $d$",
        ylabel=r"Force $F$",
        ax=None,
        label=None,
        **kwargs,
    ):
        "Create a force-displacement plot."

        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()

        if label is None:
            label = self.label

        ax.plot(self.displacement, self.force, *args, label=label, **kwargs)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        return fig, ax

    def plot_stress_stretch(
        self,
        *args,
        ax=None,
        xlabel=r"Stretch $l/L$",
        ylabel=r"Force per undeformed area $F/A$",
        label=None,
        **kwargs,
    ):
        "Create a stress-stretch plot."

        if ax is None:
            fig, ax = plt.subplots()
        else:
            fig = ax.get_figure()

        if label is None:
            label = self.label

        ax.plot(self.stretch, self.stress(), *args, label=label, **kwargs)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        return fig, ax



import felupe as fem

import hyperelastic as hel


def fea(umat):
    mesh = fem.Cube(n=3)
    region = fem.RegionHexahedron(mesh)
    field = fem.FieldContainer([fem.Field(region, dim=3)])
    boundaries, loadcase = fem.dof.uniaxial(field, clamped=True)
    solid = fem.SolidBodyNearlyIncompressible(umat, field, bulk=5000)
    move = fem.math.linsteps([0, 2], num=3)
    step = fem.Step(
        items=[solid], ramp={boundaries["move"]: move}, boundaries=boundaries
    )
    job = fem.Job(steps=[step])

    return job


def test_distortional_stretches():
    model = hel.models.stretches.Ogden(mu=[1], alpha=[0.436])
    umat = hel.DistortionalSpace(hel.StretchesFramework(model))
    fea(umat).evaluate(verbose=2)


def test_distortional_invariants():
    model = hel.models.invariants.ThirdOrderDeformation(
        C10=0.4, C01=0.1, C11=0.02, C20=-0.05, C30=0.01
    )
    umat = hel.DistortionalSpace(hel.InvariantsFramework(model))
    fea(umat).evaluate(verbose=2)


def test_distortional_generalized_strain_invariants():
    tod = hel.models.invariants.ThirdOrderDeformation(
        C10=0.4, C01=0.1, C20=0.1, strain=True
    )
    fun = hel.models.generalized.strain
    framework = hel.GeneralizedInvariantsFramework(tod, fun=fun, exponent=1.4)
    umat = hel.DistortionalSpace(framework)
    fea(umat).evaluate(verbose=2)


def test_distortional_generalized_deformation_invariants():
    tod = hel.models.invariants.ThirdOrderDeformation(
        C10=0.4, C01=0.1, C20=0.1, strain=False
    )
    fun = hel.models.generalized.deformation
    framework = hel.GeneralizedInvariantsFramework(tod, fun=fun, exponent=1.4)
    umat = hel.DistortionalSpace(framework)
    fea(umat).evaluate(verbose=2)


def test_dilatational_invariants():
    model = hel.models.invariants.ThirdOrderDeformation(C10=0.5)
    umat = hel.DilatationalSpace(hel.InvariantsFramework(model))
    fea(umat).evaluate(verbose=2)


def test_deformation_invariants():
    model = hel.models.invariants.ThirdOrderDeformation(
        C10=0.4, C01=0.1, C11=0.02, C20=-0.05, C30=0.01
    )
    umat = hel.DeformationSpace(hel.InvariantsFramework(model))
    fea(umat).evaluate(verbose=2)


if __name__ == "__main__":
    test_distortional_stretches()
    test_distortional_invariants()
    test_dilatational_invariants()
    test_deformation_invariants()
    test_distortional_generalized_strain_invariants()
    test_distortional_generalized_deformation_invariants()



import numpy as np


class Ogden:
    r"""Ogden isotropic hyperelastic material formulation based on the principal
    stretches. The strain energy density per unit undeformed volume is given as a sum of
    individual contributions from the principal stretches.

    ..  math::

        \psi(\lambda_\alpha) &= \sum_\alpha \psi_\alpha(\lambda_\alpha)

        \psi_\alpha(\lambda_\alpha) &= \frac{2 \mu}{k^2} \left(
            \lambda_\alpha^k - 1 \right)

    The first partial derivatives of the strain energy density w.r.t. the
    principal stretches are required for the principal values of the stress.

    ..  math::

        \frac{\partial \psi}{\partial \lambda_\alpha} = \sum_\alpha
            \frac{2 \mu}{k} \lambda_\alpha^{k - 1}

    Furthermore, the second partial derivatives of the strain energy density w.r.t. the
    principal stretches, necessary for the principal components of the elastic tangent
    moduli, are carried out.

    ..  math::

        \frac{\partial^2 \psi}{\partial \lambda_\alpha~\partial \lambda_\alpha} =
            \sum_\alpha \frac{2 \mu (k-1)}{k} \lambda_\alpha^{k - 2}

    """

    def __init__(self, mu, alpha):
        """Initialize the Third Order Deformation material formulation with its
        parameters.
        """

        self.mu = mu
        self.alpha = alpha

    def gradient(self, stretches, statevars):
        """The gradient as the partial derivative of the strain energy function w.r.t.
        the principal stretches."""

        dWdλα = np.zeros_like(stretches)

        for mu, alpha in zip(self.mu, self.alpha):
            dWdλα += 2 * mu / alpha * stretches ** (alpha - 1)

        return dWdλα, statevars

    def hessian(self, stretches, statevars):
        """The hessian as the second partial derivatives of the strain energy function
        w.r.t. the principal stretches."""

        d2Wdλαdλα = np.zeros_like(stretches)

        for mu, alpha in zip(self.mu, self.alpha):
            d2Wdλαdλα += 2 * mu / alpha * (alpha - 1) * stretches ** (alpha - 2)

        return [*d2Wdλαdλα, None, None, None]



import numpy as np

from ..math import astensor, asvoigt, cdya_ik, eye


class Deformation:
    r"""The deformation space.

    This class takes a Total-Lagrange material formulation and applies it on the
    deformation space.

    ..  math::

        \psi = \psi(\boldsymbol{C}(\boldsymbol{F}))

    The gradient of the strain energy function is carried out w.r.t. the Green Lagrange
    strain tensor. Hence, the work-conjugate stress tensor here refers to the second
    Piola-Kirchhoff stress tensor.

    ..  math::

        \boldsymbol{S} = \frac{\partial \psi}{\partial \frac{1}{2}\boldsymbol{C}}

    The hessian of the strain energy function is carried out w.r.t. the Green-Lagrange
    strain tensor. Hence, the work-conjugate elasticity tensor here refers to the
    fourth-order Total-Lagrangian elasticity tensor.

    ..  math::

        \mathbb{C} = \frac{\partial^2 \psi}{\partial \frac{1}{2}\boldsymbol{C}~
            \frac{1}{2}\boldsymbol{C}}


    Given a Total-Lagrange material formulation, for the variation and linearization of
    the virtual work of internal forces, the output quantities have to be transformed:
    The second Piola-Kirchhoff stress tensor is converted into the deformation gradient
    work-conjugate first Piola-Kirchhoff stress tensor, along with its fourth-order
    elasticity tensor. Also, the so-called geometric tangent stiffness component
    (initial stress matrix) is added to the fourth-order elasticity tensor.

    ..  math::

        \delta W_{int} &=
            - \int_V \boldsymbol{P} : \delta \boldsymbol{F} ~ dV

        \Delta \delta W_{int} &=
            - \int_V \delta \boldsymbol{F} : \mathbb{A} : \Delta \boldsymbol{F} ~ dV

    where

    ..  math::

        \boldsymbol{P} &= \boldsymbol{F} \boldsymbol{S}

        \mathbb{A}_{iJkL} &= F_{iI} F_{kK} \mathbb{C}_{IJKL} + \delta_{ik} S_{JL}

    """

    def __init__(self, material, parallel=False):
        self.parallel = parallel
        self.material = material

        if self.parallel:
            from einsumt import einsumt

            self.einsum = einsumt
        else:
            self.einsum = np.einsum

        # initial variables for calling
        # ``self.gradient(self.x)`` and ``self.hessian(self.x)``
        self.x = [material.x[0], material.x[-1]]

    def gradient(self, x):
        """The gradient as the partial derivative of the strain energy function w.r.t.
        the deformation gradient."""

        F, statevars = x[0], x[-1]

        self.C = asvoigt(self.einsum("kI...,kJ...->IJ...", F, F))
        dWdC, statevars_new = self.material.gradient(self.C, statevars)

        self.S = 2 * dWdC

        return [self.einsum("iK...,KJ...->iJ...", F, astensor(self.S)), statevars_new]

    def hessian(self, x):
        """The hessian as the second partial derivative of the strain energy function
        w.r.t. the deformation gradient."""

        F, statevars = x[0], x[-1]

        dWdF, statevars_new = self.gradient(x)
        I = eye(self.C)

        d2WdCdC = self.material.hessian(self.C, statevars)
        C4 = 4 * d2WdCdC

        if self.parallel:
            from einsumt import einsumt

            einsum = einsumt
        else:
            einsum = np.einsum

        A4 = einsum("iI...,kK...,IJKL...->iJkL...", F, F, astensor(C4, 4))

        return [A4 + cdya_ik(I, self.S)]



import numpy as np

from ._plotting import LabPlotter


class Simulation(LabPlotter):
    """Results of a simulation along with methods to convert and plot the data.

    Attributes
    ----------
    loadcase : class
        A class with methods for evaluating the deformation gradient and the stress
        as normal force per undeformed area, e.g. :class:`Uniaxial <.lab.Uniaxial>`.
    stretch : ndarray
        The stretch as the ratio of the deformed vs. the undeformed length.
    labels : list of str
        A list of the material parameter labels.
    material : class
        A class with a method for evaluating the gradient of the strain energy function
        w.r.t. the deformation gradient, e.g.
        :class:`DistortionalSpace <.DistortionalSpace>`.
    parameters : array_like
        The material parameters.


    Examples
    --------
    The material model response behaviour of a hyperelastic material model formulation
    is evaluated for a :class:`uniaxial tension <.lab.Uniaxial>` load case. A given
    stretch data is used to create the :class:`Simulation <.lab.Simulation>` object.

    >>> import numpy as np
    >>> import hyperelastic
    >>> from hyperelastic import lab
    >>>
    >>> stretch = np.linspace(0.7, 2.5, 181)

    A function which takes the material parameters and returns the hyperelastic
    constitutive material formulation has to be provided for the simulation objects.
    Here, we use an isotropic invariant-based
    :class:`third-order deformation <.models.invariants.ThirdOrderDeformation>`
    material formulation.

    >>> def material(**kwargs):
    >>>     "A third-order deformation material formulation."
    >>>
    >>>     tod = hyperelastic.models.invariants.ThirdOrderDeformation(**kwargs)
    >>>     framework = hyperelastic.InvariantsFramework(tod)
    >>>
    >>>     return hyperelastic.DeformationSpace(framework)

    A list of (string) labels is used to apply the list or array of parameter values to
    the material formulation.

    >>> simulation = lab.Simulation(
    >>>     loadcase=lab.Uniaxial(),
    >>>     stretch=stretch,
    >>>     material=material,
    >>>     labels=["C10", "C01", "C11", "C20", "C30"],
    >>>     parameters=[0.4, 0.1, 0.02, -0.04, 0.01],
    >>> )

    The stress-stretch plot returns a figure which visualizes the force per undeformed
    area vs. the ratio of the undeformed and deformed length.

    >>> fig, ax = simulation.plot_stress_stretch()
    >>>
    >>> ax.legend()
    >>> ax.set_title("Third-Order Deformation")

    ..  image:: images/fig_simulation-tod.png

    """

    def __init__(self, loadcase, stretch, labels, material, parameters=None):
        """Initialize the results of a simulation.

        Parameters
        ----------
        loadcase : class
            A class with methods for evaluating the deformation gradient and the stress
            as normal force per undeformed area, e.g.
            :class:`Uniaxial <.lab.Uniaxial>`.
        stretch : ndarray
            The stretch as the ratio of the deformed vs. the undeformed length.
        labels : list of str
            A list of the material parameter labels.
        material : class
            A class with a method for evaluating the gradient of the strain energy
            function w.r.t. the deformation gradient, e.g.
            :class:`DistortionalSpace <.DistortionalSpace>`.
        parameters : array_like or None, optional
            The material parameters (default is None).

        """

        super().__init__()

        self.loadcase = loadcase
        self.stretch = stretch
        self.labels = labels
        self.parameters = parameters
        self.material = material

        self.label = loadcase.label

        if self.parameters is None:
            self.parameters = np.ones(len(self.labels))

    def stress_curve_fit(self, x, *parameters):
        """Evaluate the stress as force per undeformed area for given material
        parameters."""

        kwargs = {label: p for label, p in zip(self.labels, parameters)}
        mat = self.material(**kwargs)
        F = self.loadcase.defgrad(x)
        P = mat.gradient([F.reshape(3, 3, 1, -1), None])[0].reshape(3, 3, -1)
        return self.loadcase.stress(F, P)

    def stress(self):
        "Evaluate the stress as force per undeformed area."

        return self.stress_curve_fit(self.stretch, *self.parameters)



import numpy as np

import hyperelastic


def test_lab_simple():
    area = 25
    length = 100

    displacement = np.linspace(0, 2 * length, 100)
    stretch = 1 + displacement / length
    force = (stretch - 1 / stretch**2 + (stretch - 1) ** 5 / 10) * area

    def material(**kwargs):
        "A third-order deformation material formulation."

        tod = hyperelastic.models.invariants.ThirdOrderDeformation(**kwargs)
        framework = hyperelastic.InvariantsFramework(tod)

        return hyperelastic.DeformationSpace(framework)

    experiments = [
        hyperelastic.lab.Experiment(
            label="Uniaxial Tension",
            displacement=displacement,
            force=force,
            area=area,
            length=length,
        ),
        hyperelastic.lab.Experiment(
            label="Planar Tension",
            displacement=displacement[::2],
            force=force[::2],
            area=area / (8 / 7),
            length=length,
        ),
        hyperelastic.lab.Experiment(
            label="Biaxial Tension",
            displacement=displacement[::2] / 2,
            force=force[::2],
            area=area / (4 / 5),
            length=length,
        ),
    ]

    labels = ["C10", "C01", "C11", "C20", "C30"]
    simulations = [
        hyperelastic.lab.Simulation(
            loadcase=hyperelastic.lab.Uniaxial(),
            stretch=experiments[0].stretch,
            material=material,
            labels=labels,
        ),
        hyperelastic.lab.Simulation(
            loadcase=hyperelastic.lab.Planar(),
            stretch=experiments[1].stretch,
            material=material,
            labels=labels,
        ),
        hyperelastic.lab.Simulation(
            loadcase=hyperelastic.lab.Biaxial(),
            stretch=experiments[2].stretch,
            material=material,
            labels=labels,
        ),
    ]

    optimize = hyperelastic.lab.Optimize(
        experiments=experiments,
        simulations=simulations,
        parameters=np.ones(5),
    )

    parameters, pcov = optimize.curve_fit(method="lm")
    fig, ax = optimize.plot(title="Third-Order Deformation")

    return fig, parameters


if __name__ == "__main__":
    fig, parameters = test_lab_simple()
    print(parameters)
    fig.savefig("../docs/hyperelastic/images/fig_optimize-tod.png")



from . import frameworks, lab, math, models, spaces
from .__about__ import __version__
from .frameworks import GeneralizedInvariants as GeneralizedInvariantsFramework
from .frameworks import Invariants as InvariantsFramework
from .frameworks import Stretches as StretchesFramework
from .spaces import Deformation as DeformationSpace
from .spaces import Dilatational as DilatationalSpace
from .spaces import Distortional as DistortionalSpace

__all__ = [
    "lab",
    "spaces",
    "frameworks",
    "math",
    "models",
    "__version__",
    "DistortionalSpace",
    "DilatationalSpace",
    "DeformationSpace",
    "InvariantsFramework",
    "StretchesFramework",
    "GeneralizedInvariantsFramework",
]



import numpy as np

from ..math import astensor, asvoigt, cdya, cdya_ik, ddot, det, dya, eye, inv, transpose


class Distortional:
    r"""The distortional (part of the deformation) space is a partial deformation with
    constant volume. For a given deformation map :math:`\boldsymbol{x}(\boldsymbol{X})`
    and its deformation gradient :math:`\boldsymbol{F}`, the distortional part of the
    deformation gradient :math:`\hat{\boldsymbol{F}}` is obtained by a multiplicative
    (consecutive) split into a volume-changing (dilatational) and a constant-volume
    (distortional) part of the deformation gradient. Due to the fact that the
    dilatational part is proportional to the unit tensor, the order of these partial
    deformations is not unique.

    ..  math::

        \boldsymbol{F} = \overset{\circ}{\boldsymbol{F}} \hat{\boldsymbol{F}} =
            \hat{\boldsymbol{F}} \overset{\circ}{\boldsymbol{F}}

    This class takes a Total-Lagrange material formulation and applies it only on the
    distortional space.

    ..  math::

        \hat{\psi} = \psi(\hat{\boldsymbol{C}}(\boldsymbol{F}))

    The distortional (unimodular) part of the right Cauchy-Green deformation tensor is
    evaluated by the help of its third invariant (the determinant). The determinant of
    a distortional (an unimodular) tensor equals to one.

    ..  math::

        \hat{\boldsymbol{C}} = I_3^{-1/3} \boldsymbol{C}

    The gradient of the strain energy function is carried out w.r.t. the Green Lagrange
    strain tensor. Hence, the work-conjugate stress tensor used in this space
    projection refers to the second Piola-Kirchhoff stress tensor.

    ..  math::

        \boldsymbol{S}' = \frac{\partial \hat{\psi}}{\partial \frac{1}{2}\boldsymbol{C}}

    The distortional space projection leads to a **physically** deviatoric second
    Piola-Kirchhoff stress tensor, evaluated by the application of the chain rule.

    ..  math::

        \hat{\boldsymbol{S}} = \frac{\partial \hat{\psi}}
            {\partial \frac{1}{2}\hat{\boldsymbol{C}}}

    The (**phyiscally**) deviatoric projection is obtained by the partial derivative of
    the distortional part of the right Cauchy-Green deformation tensor w.r.t. the right
    Cauchy-Green deformation tensor.

    ..  math::

        \frac{\partial \hat{\boldsymbol{C}}}{\partial \boldsymbol{C}} =
            \frac{\partial I_3^{-1/3} \boldsymbol{C}}{\partial \boldsymbol{C}} =
            I_3^{-1/3} \left( \boldsymbol{1} \odot \boldsymbol{1}
            - \frac{1}{3} \boldsymbol{C} \otimes \boldsymbol{C}^{-1} \right)

    This partial derivative is used to perform the distortional space projection of the
    second Piola-Kirchhoff stress tensor. Instead of asserting the determinant-scaling
    to the fourth-order projection tensor, this factor is combined with the second
    Piola-Kirchhoff stress tensor in the distortional space. Hence, the stress tensor in
    the distortional space, scaled by :math:`I_3^{-1/3}`, is introduced as a new
    (frequently re-used) variable, denoted by an overset bar.

    ..  math::

        \boldsymbol{S}' &= \mathbb{P} : \bar{\boldsymbol{S}}

        \bar{\boldsymbol{S}} &= I_3^{-1/3} \hat{\boldsymbol{S}}

        \mathbb{P} &= \boldsymbol{1} \odot \boldsymbol{1}
            - \frac{1}{3} \boldsymbol{C}^{-1} \otimes \boldsymbol{C}

    The evaluation of the double-dot product for the distortional space projection leads
    to the mathematical deviator of the product between the scaled distortional space
    stress tensor and the right Cauchy-Green deformation tensor, right multiplied by the
    inverse of the right Cauchy-Green deformation tensor.

    ..  math::

        \boldsymbol{S}' = \bar{\boldsymbol{S}}
            - \frac{\bar{\boldsymbol{S}}:\boldsymbol{C}}{3} \boldsymbol{C}^{-1}
            = \text{dev}(\bar{\boldsymbol{S}} \boldsymbol{C}) \boldsymbol{C}^{-1}

    The hessian of the strain energy function is carried out w.r.t. the Green-Lagrange
    strain tensor. Hence, the work-conjugate elasticity tensor used in this space
    projection refers to the fourth-order Total-Lagrangian elasticity tensor.

    ..  math::

        \mathbb{C}' = \frac{\partial^2 \hat{\psi}}{\partial \frac{1}{2}\boldsymbol{C}~
            \frac{1}{2}\boldsymbol{C}}

    The evaluation of this second partial derivative leads to the elasticity tensor of
    the distortional space projection. The remaining determinant scaling terms of the
    projection tensor are included in the determinant-modified fourth-order elasticity
    tensor, denoted with an overset bar.

    ..  math::

        \mathbb{C}' = \mathbb{P} : \bar{\mathbb{C}} : \mathbb{P}^T + \frac{2}{3} \left(
            \left( \bar{\boldsymbol{S}}:\boldsymbol{C} \right)
            \boldsymbol{C}^{-1} \odot \boldsymbol{C}^{-1}
            - \bar{\boldsymbol{S}} \otimes \boldsymbol{C}^{-1}
            - \boldsymbol{C}^{-1} \otimes \bar{\boldsymbol{S}}
            + \frac{1}{3} \left( \bar{\boldsymbol{S}}:\boldsymbol{C} \right)
            \boldsymbol{C}^{-1} \otimes \boldsymbol{C}^{-1}
        \right)

    ..  math::

        \bar{\mathbb{C}} = I_3^{-2/3} \hat{\mathbb{C}}

    For the variation and linearization of the virtual work of internal forces, the
    output quantities have to be transformed: The second Piola-Kirchhoff stress tensor
    is converted into the  deformation gradient work-conjugate first Piola-Kirchhoff
    stress tensor, along with its fourth-order elasticity tensor. Also, the so-called
    geometric tangent stiffness component (initial stress matrix) is added to the
    fourth-order elasticity tensor.

    ..  math::

        \delta W_{int} &=
            - \int_V \boldsymbol{P} : \delta \boldsymbol{F} ~ dV

        \Delta \delta W_{int} &=
            - \int_V \delta \boldsymbol{F} : \mathbb{A} : \Delta \boldsymbol{F} ~ dV

    where

    ..  math::

        \boldsymbol{P} &= \boldsymbol{F} \boldsymbol{S}

        \mathbb{A}_{iJkL} &= F_{iI} F_{kK} \mathbb{C}_{IJKL} + \delta_{ik} S_{JL}

    """

    def __init__(self, material, parallel=False):
        self.parallel = parallel
        self.material = material

        if self.parallel:
            from einsumt import einsumt

            self.einsum = einsumt
        else:
            self.einsum = np.einsum

        # initial variables for calling
        # ``self.gradient(self.x)`` and ``self.hessian(self.x)``
        self.x = [material.x[0], material.x[-1]]

    def gradient(self, x):
        """The gradient as the partial derivative of the strain energy function w.r.t.
        the deformation gradient."""

        F, statevars = x[0], x[-1]

        self.C = C = asvoigt(self.einsum("kI...,kJ...->IJ...", F, F))
        self.I3 = I3 = det(C)
        self.Cu = I3 ** (-1 / 3) * C
        dWudCu, statevars_new = self.material.gradient(self.Cu, statevars)

        self.Sb = Sb = 2 * I3 ** (-1 / 3) * dWudCu
        self.SbC = SbC = ddot(Sb, C)
        self.invC = invC = inv(C, determinant=I3)

        self.S = Sb - SbC / 3 * invC

        return [self.einsum("iK...,KJ...->iJ...", F, astensor(self.S)), statevars_new]

    def hessian(self, x):
        """The hessian as the second partial derivative of the strain energy function
        w.r.t. the deformation gradient."""

        F, statevars = x[0], x[-1]

        dWudF, statevars_new = self.gradient(x)
        I = eye(self.C)

        d2WdCdC = self.material.hessian(self.Cu, statevars)
        C4b = 4 * self.I3 ** (-2 / 3) * d2WdCdC

        P4 = cdya(I, I) - dya(self.invC, self.C) / 3
        I4 = cdya(self.invC, self.invC)

        SbinvC = dya(self.Sb, self.invC)
        invCSb = transpose(SbinvC)
        invCinvC = dya(self.invC, self.invC)

        C4 = 2 / 3 * (self.SbC * I4 - SbinvC - invCSb + self.SbC / 3 * invCinvC)

        if not np.allclose(C4b, 0):
            C4 += ddot(ddot(P4, C4b, mode=(4, 4)), transpose(P4), mode=(4, 4))

        A4 = self.einsum("iI...,kK...,IJKL...->iJkL...", F, F, astensor(C4, 4))

        return [A4 + cdya_ik(I, self.S)]



from functools import wraps

import numpy as np
from scipy.optimize import curve_fit


@wraps(curve_fit)
def concatenate_curve_fit(
    f,
    xdata,
    ydata,
    *args,
    **kwargs,
):
    """Concatenate lists of functions, xdata and ydata and apply them on
    ``scipy.optimize.curve_fit()``."""

    def concatenate_f(x, *p):
        "Evaluate and return a concatenated array for a list of functions."
        return np.concatenate([fi(xi, *p) for fi, xi in zip(f, xdata)])

    return curve_fit(
        concatenate_f,
        np.concatenate(xdata),
        np.concatenate(ydata),
        *args,
        **kwargs,
    )



import felupe as fem
import sympy as sym

import hyperelastic as hel


def fea(umat):
    mesh = fem.Cube(n=3)
    region = fem.RegionHexahedron(mesh)
    field = fem.FieldContainer([fem.Field(region, dim=3)])
    boundaries, loadcase = fem.dof.uniaxial(field, clamped=True)
    solid = fem.SolidBodyNearlyIncompressible(umat, field, bulk=5000)
    move = fem.math.linsteps([0, 2], num=3)
    step = fem.Step(
        items=[solid], ramp={boundaries["move"]: move}, boundaries=boundaries
    )
    job = fem.Job(steps=[step])

    return job


def ogden(stretches, mu, alpha):
    x, y, z = stretches
    return 2 * mu / alpha**2 * (x**alpha + y**alpha + z**alpha - 3)


class SymbolicMaterialFormulation:
    def __init__(self, strain_energy_function, **kwargs):
        self.x, self.y, self.z = sym.symbols("λ_1 λ_2 λ_3")
        self.W = strain_energy_function([self.x, self.y, self.z], **kwargs)

        self.dWdx = sym.lambdify([self.x, self.y, self.z], self.W.diff(self.x))
        self.dWdy = sym.lambdify([self.x, self.y, self.z], self.W.diff(self.y))
        self.dWdz = sym.lambdify([self.x, self.y, self.z], self.W.diff(self.z))

        self.d2Wdxdx = sym.lambdify(
            [self.x, self.y, self.z], self.W.diff(self.x, self.x)
        )
        self.d2Wdydy = sym.lambdify(
            [self.x, self.y, self.z], self.W.diff(self.y, self.y)
        )
        self.d2Wdzdz = sym.lambdify(
            [self.x, self.y, self.z], self.W.diff(self.z, self.z)
        )

        self.d2Wdxdy = sym.lambdify(
            [self.x, self.y, self.z], self.W.diff(self.x, self.y)
        )
        self.d2Wdydz = sym.lambdify(
            [self.x, self.y, self.z], self.W.diff(self.y, self.z)
        )
        self.d2Wdxdz = sym.lambdify(
            [self.x, self.y, self.z], self.W.diff(self.x, self.z)
        )

    def gradient(self, stretches, statevars):
        return [
            self.dWdx(*stretches),
            self.dWdy(*stretches),
            self.dWdz(*stretches),
        ], statevars

    def hessian(self, stretches, statevars):
        return [
            self.d2Wdxdx(*stretches),
            self.d2Wdydy(*stretches),
            self.d2Wdzdz(*stretches),
            self.d2Wdxdy(*stretches),
            self.d2Wdydz(*stretches),
            self.d2Wdxdz(*stretches),
        ]


def test_distortional_stretches_sympy():
    model = SymbolicMaterialFormulation(ogden, mu=1, alpha=0.436)
    umat = hel.DistortionalSpace(hel.StretchesFramework(model))
    fea(umat).evaluate(verbose=2)


if __name__ == "__main__":
    test_distortional_stretches_sympy()



import numpy as np

from ..math import cdya, ddot, det, dya, eye, inv, trace


class Invariants:
    r"""The Framework for a Total-Lagrangian invariant-based isotropic hyperelastic
    material formulation provides the material behaviour-independent parts for
    evaluating the second Piola-Kirchhoff stress tensor as well as its associated
    fourth-order elasticity tensor.

    The gradient as well as the hessian of the strain energy function are carried out
    w.r.t. the right Cauchy-Green deformation tensor. Hence, the work-conjugate stress
    tensor is one half of the second Piola-Kirchhoff stress tensor and the fourth-order
    elasticitiy tensor used here is a quarter of the Total-Lagrangian elasticity tensor.

    ..  math::

        \psi(\boldsymbol{C}) =
            \psi(I_1(\boldsymbol{C}), I_2(\boldsymbol{C}), I_3(\boldsymbol{C}))

    The first and second invariants of the left or right Cauchy-Green deformation tensor
    are identified as factors of their characteristic polynomial,

    ..  math::

        I_1 &= \text{tr}(\boldsymbol{C})

        I_2 &= \frac{1}{2}
            \left( \text{tr}(\boldsymbol{C})^2 - \text{tr}(\boldsymbol{C}^2) \right)

        I_3 &= \det(\boldsymbol{C})

    where the Cauchy-Green deformation tensors eliminate the rigid body rotations
    of the deformation gradient and serve as a quadratic change-of-length measure of
    the deformation.

    ..  math::

        \boldsymbol{C} &= \boldsymbol{F}^T \boldsymbol{F}

        \boldsymbol{b} &= \boldsymbol{F} \boldsymbol{F}^T

    The first partial derivatives of the strain energy function w.r.t. the invariants

    ..  math::

        \psi_{,1} &= \frac{\partial \psi}{\partial I_1}

        \psi_{,2} &= \frac{\partial \psi}{\partial I_2}

        \psi_{,3} &= \frac{\partial \psi}{\partial I_3}

    and the partial derivatives of the invariants w.r.t. the right Cauchy-Green
    deformation tensor are defined.

    ..  math::

        \frac{\partial I_1}{\partial \boldsymbol{C}} &= \boldsymbol{I}

        \frac{\partial I_2}{\partial \boldsymbol{C}} &=
            \left( I_1 \boldsymbol{I} - \boldsymbol{C} \right)

        \frac{\partial I_3}{\partial \boldsymbol{C}} &= I_3 \boldsymbol{C}^{-1}

    The second Piola-Kirchhoff stress tensor is formulated by the application of the
    chain rule.

    ..  math::

        \frac{\partial \psi}{\partial \boldsymbol{C}} =
            \frac{\partial \psi}{\partial I_1}
            \frac{\partial I_1}{\partial \boldsymbol{C}}
            + \frac{\partial \psi}{\partial I_2}
            \frac{\partial I_2}{\partial \boldsymbol{C}}
            + \frac{\partial \psi}{\partial I_3}
            \frac{\partial I_3}{\partial \boldsymbol{C}}

    Furthermore, the second partial derivatives of the elasticity tensor are carried
    out.

    ..  math::

        \frac{\partial^2 \psi}{\partial \boldsymbol{C}~\partial \boldsymbol{C}} &=
            \frac{\partial^2 \psi}{\partial I_1~\partial I_1}
            \left( \frac{\partial I_1}{\partial \boldsymbol{C}} \otimes
            \frac{\partial I_1}{\partial \boldsymbol{C}} \right)

            &+ \frac{\partial^2 \psi}{\partial I_2~\partial I_2}
            \left( \frac{\partial I_2}{\partial \boldsymbol{C}} \otimes
            \frac{\partial I_2}{\partial \boldsymbol{C}} \right)

            &+ \frac{\partial^2 \psi}{\partial I_3~\partial I_3}
            \left( \frac{\partial I_3}{\partial \boldsymbol{C}} \otimes
            \frac{\partial I_3}{\partial \boldsymbol{C}} \right)

            &+ \frac{\partial^2 \psi}{\partial I_1~\partial I_2}
            \left( \frac{\partial I_1}{\partial \boldsymbol{C}} \otimes
            \frac{\partial I_2}{\partial \boldsymbol{C}}
            + \frac{\partial I_2}{\partial \boldsymbol{C}} \otimes
            \frac{\partial I_1}{\partial \boldsymbol{C}} \right)

            &+ \frac{\partial^2 \psi}{\partial I_2~\partial I_3}
            \left( \frac{\partial I_2}{\partial \boldsymbol{C}} \otimes
            \frac{\partial I_3}{\partial \boldsymbol{C}}
            + \frac{\partial I_3}{\partial \boldsymbol{C}} \otimes
            \frac{\partial I_2}{\partial \boldsymbol{C}} \right)

            &+ \frac{\partial^2 \psi}{\partial I_1~\partial I_3}
            \left( \frac{\partial I_1}{\partial \boldsymbol{C}} \otimes
            \frac{\partial I_3}{\partial \boldsymbol{C}}
            + \frac{\partial I_3}{\partial \boldsymbol{C}} \otimes
            \frac{\partial I_1}{\partial \boldsymbol{C}} \right)

            &+ \frac{\partial \psi}{\partial I_1}
            \frac{\partial^2 I_1}{\partial \boldsymbol{C}~\partial \boldsymbol{C}}
            + \frac{\partial \psi}{\partial I_2}
            \frac{\partial^2 I_2}{\partial \boldsymbol{C}~\partial \boldsymbol{C}}
            + \frac{\partial \psi}{\partial I_3}
            \frac{\partial^2 I_3}{\partial \boldsymbol{C}~\partial \boldsymbol{C}}

    The only non material behaviour-related terms which are not already defined during
    stress evaluation are the second partial derivatives of the invariants w.r.t.
    the right Cauchy-Green deformation tensor.

    ..  math::

        \frac{\partial^2 I_1}{\partial \boldsymbol{C}~\partial \boldsymbol{C}} &=
            \mathbb{0}

        \frac{\partial^2 I_2}{\partial \boldsymbol{C}~\partial \boldsymbol{C}} &=
            \boldsymbol{I} \otimes \boldsymbol{I} - \boldsymbol{I} \odot \boldsymbol{I}

        \frac{\partial^2 I_3}{\partial \boldsymbol{C}~\partial \boldsymbol{C}} &=
            I_3 \left( \boldsymbol{C}^{-1} \otimes \boldsymbol{C}^{-1}
                - \boldsymbol{C}^{-1} \odot \boldsymbol{C}^{-1}
            \right)

    """

    def __init__(self, material, nstatevars=0, parallel=False):
        """Initialize the Framework for an invariant-based isotropic hyperelastic
        material formulation."""

        self.parallel = parallel
        self.material = material

        self.x = [np.eye(3), np.zeros(nstatevars)]

    def gradient(self, C, statevars):
        """The gradient as the partial derivative of the strain energy function w.r.t.
        the right Cauchy-Green deformation tensor (one half of the second Piola
        Kirchhoff stress tensor)."""

        self.I = I = eye(C)

        self.I1 = I1 = trace(C)
        self.I2 = (I1**2 - ddot(C, C)) / 2
        self.I3 = I3 = det(C)

        self.dWdI1, self.dWdI2, self.dWdI3, statevars_new = self.material.gradient(
            self.I1, self.I2, self.I3, statevars
        )

        ntrax = len(C.shape[1:])
        dWdC = np.zeros((6, *np.ones(ntrax, dtype=int)))

        if self.dWdI1 is not None:
            self.dI1dC = I
            dWdC = dWdC + self.dWdI1 * self.dI1dC

        if self.dWdI2 is not None:
            self.dI2dC = I1 * I - C
            dWdC = dWdC + self.dWdI2 * self.dI2dC

        if self.dWdI3 is not None:
            self.invC = invC = inv(C, determinant=self.I3)
            self.dI3dC = I3 * invC
            dWdC = dWdC + self.dWdI3 * self.dI3dC

        return dWdC, statevars_new

    def hessian(self, C, statevars):
        """The hessian as the second partial derivatives of the strain energy function
        w.r.t. the right Cauchy-Green deformation tensor (a quarter of the Lagrangian
        fourth-order elasticity tensor associated to the second Piola-Kirchhoff stress
        tensor)."""

        dWdC, statevars = self.gradient(C, statevars)
        (
            d2WdI1dI1,
            d2WdI2dI2,
            d2WdI3dI3,
            d2WdI1dI2,
            d2WdI2dI3,
            d2WdI1dI3,
        ) = self.material.hessian(self.I1, self.I2, self.I3, statevars)
        I = self.I

        ntrax = len(C.shape[1:])
        d2WdCdC = np.zeros((6, 6, *np.ones(ntrax, dtype=int)))

        if self.dWdI1 is not None:
            dI1dC = self.dI1dC
            # d2I1dCdC = 0
            # d2WdCdC = d2WdCdC + self.dWdI1 * d2I1dCdC

        if self.dWdI2 is not None:
            dI2dC = self.dI2dC
            d2I2dCdC = dya(I, I) - cdya(I, I)
            d2WdCdC = d2WdCdC + self.dWdI2 * d2I2dCdC

        if self.dWdI3 is not None:
            dI3dC = self.dI3dC
            invC = self.invC
            d2I3dCdC = self.I3 * (dya(invC, invC) - cdya(invC, invC))
            d2WdCdC = d2WdCdC + self.dWdI3 * d2I3dCdC

        if d2WdI1dI1 is not None:
            d2WdCdC = d2WdCdC + d2WdI1dI1 * dya(dI1dC, dI1dC)

        if d2WdI2dI2 is not None:
            d2WdCdC = d2WdCdC + d2WdI2dI2 * dya(dI2dC, dI2dC)

        if d2WdI3dI3 is not None:
            d2WdCdC = d2WdCdC + d2WdI2dI2 * dya(dI2dC, dI2dC)

        if d2WdI1dI2 is not None:
            d2WdCdC = d2WdCdC + d2WdI1dI2 * (dya(dI1dC, dI2dC) + dya(dI2dC, dI1dC))

        if d2WdI2dI3 is not None:
            d2WdCdC = d2WdCdC + d2WdI2dI3 * (dya(dI2dC, dI3dC) + dya(dI3dC, dI2dC))

        if d2WdI1dI3 is not None:
            d2WdCdC = d2WdCdC + d2WdI1dI3 * (dya(dI1dC, dI3dC) + dya(dI3dC, dI1dC))

        return d2WdCdC



import numpy as np

import hyperelastic.math as hm


def test_math():
    np.random.seed(546123)
    F = np.eye(3).reshape(3, 3, 1, 1) + np.random.rand(3, 3, 4, 10) / 10
    C = np.einsum("ki...,kj...->ij...", F, F)
    D = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3, 1, 1) * C
    A = np.einsum("ij...,kl...->ijkl...", C, C)
    C6 = hm.asvoigt(C, mode=2)
    D6 = hm.asvoigt(D, mode=2)
    A6 = hm.asvoigt(A, mode=4)

    assert np.allclose(hm.tril_from_triu(A6), A6)
    assert np.allclose(hm.triu_from_tril(A6), A6)
    assert np.allclose(hm.trace(hm.dev(C6)), 0)

    assert np.allclose(hm.inv(C6), hm.inv(C6, determinant=hm.det(C6)))
    assert np.allclose(hm.dot(C6, C6), np.einsum("ik...,kj...->ij...", C, C))

    assert np.allclose(hm.ddot(C6, A6, mode=(2, 4)), hm.ddot(A6, C6, mode=(4, 2)))

    assert np.allclose(hm.cdya_il(C6, C6), np.einsum("il...,kj...->ijkl...", C, C))
    assert np.allclose(hm.cdya_ik(C6, C6), np.einsum("ik...,jl...->ijkl...", C, C))

    CD = (
        np.einsum("ik...,jl...->ijkl...", C, D)
        + np.einsum("il...,kj...->ijkl...", C, D)
    ) / 2
    DC = (
        np.einsum("ik...,jl...->ijkl...", D, C)
        + np.einsum("il...,kj...->ijkl...", D, C)
    ) / 2
    B = (CD + DC) / 2

    assert np.allclose(B, hm.astensor(hm.cdya(C6, D6), mode=4))
    assert np.allclose(B, hm.astensor(hm.cdya(D6, C6), mode=4))


if __name__ == "__main__":
    test_math()



import numpy as np

from ..math import astensor, asvoigt, cdya, cdya_ik, ddot, det, dya, eye, inv, trace


class Dilatational:
    def __init__(self, material, parallel=False):
        self.parallel = parallel
        self.material = material

        if self.parallel:
            from einsumt import einsumt

            self.einsum = einsumt
        else:
            self.einsum = np.einsum

        # initial variables for calling
        # ``self.gradient(self.x)`` and ``self.hessian(self.x)``
        self.x = [material.x[0], material.x[-1]]

    def gradient(self, x):
        F, statevars = x[0], x[-1]

        self.C = asvoigt(self.einsum("kI...,kJ...->IJ...", F, F))
        self.I3 = det(self.C)
        self.Co = self.I3 ** (1 / 3) * self.C
        dWodCo, statevars_new = self.material.gradient(self.Co, statevars)

        self.Sb = 2 * self.I3 ** (1 / 3) * dWodCo
        self.tr_Sb = trace(self.Sb)
        self.invC = inv(self.C, determinant=self.I3)

        self.S = self.tr_Sb / 3 * self.invC
        return [self.einsum("iK...,KJ...->iJ...", F, astensor(self.S)), statevars_new]

    def hessian(self, x):
        F, statevars = x[0], x[-1]

        dWodF, statevars_new = self.gradient(x)

        d2WodCodCo = self.material.hessian(self.Co, statevars)
        C4b = 4 * self.I3 ** (2 / 3) * d2WodCodCo

        I = eye(self.C)
        tr_C4b = ddot(ddot(I, C4b, mode=(2, 4)), I)
        I4 = cdya(self.invC, self.invC)

        C4 = (2 * self.tr_Sb + tr_C4b) * dya(
            self.invC, self.invC
        ) / 9 - 2 / 3 * self.tr_Sb * I4

        A4 = self.einsum("iI...,kK...,IJKL...->iJkL...", F, F, astensor(C4, 4))

        return [A4 + cdya_ik(I, self.S)]



import numpy as np

from ..math import cdya, dya, eigh, transpose


class Stretches:
    r"""The Framework for a Total-Lagrangian stretch-based isotropic hyperelastic
    material formulation provides the material behaviour-independent parts for
    evaluating the second Piola-Kirchhoff stress tensor as well as its associated
    fourth-order elasticity tensor.

    The gradient as well as the hessian of the strain energy function are carried out
    w.r.t. the right Cauchy-Green deformation tensor. Hence, the work-conjugate stress
    tensor is one half of the second Piola-Kirchhoff stress tensor and the fourth-order
    elasticitiy tensor used here is a quarter of the Total-Lagrangian elasticity tensor.

    ..  math::

        \psi(\boldsymbol{C}) = \psi(\lambda_\alpha(\boldsymbol{C}))

    The principal stretches (the square roots of the eigenvalues) of the left or right
    Cauchy-Green deformation tensor are obtained by the solution of the eigenvalue
    problem,

    ..  math::

        \left( \boldsymbol{C} - \lambda^2_\alpha \boldsymbol{I} \right)
            \boldsymbol{N}_\alpha = \boldsymbol{0}

    where the Cauchy-Green deformation tensors eliminate the rigid body rotations
    of the deformation gradient and serve as a quadratic change-of-length measure of
    the deformation.

    ..  math::

        \boldsymbol{C} &= \boldsymbol{F}^T \boldsymbol{F}

        \boldsymbol{b} &= \boldsymbol{F} \boldsymbol{F}^T

    The first partial derivative of the strain energy function w.r.t. a principal
    stretch

    ..  math::

        \psi_{,\alpha} = \frac{\partial \psi}{\partial \lambda_\alpha}

    and the partial derivative of a principal stretch w.r.t. the right Cauchy-Green
    deformation tensor is defined

    ..  math::

        \frac{\partial \lambda_\alpha}{\partial \boldsymbol{C}} =
            \frac{\partial (\lambda^2_\alpha)^{1/2}}{\partial \boldsymbol{C}} =
            \frac{1}{2 \lambda_\alpha} \boldsymbol{M}_\alpha

    with the eigenbase as the dyadic (outer vector) product of eigenvectors.

    ..  math::

        \boldsymbol{M}_\alpha = \boldsymbol{N}_\alpha \otimes \boldsymbol{N}_\alpha

    The second Piola-Kirchhoff stress tensor is formulated by the application of the
    chain rule and a sum of all principal stretch contributions.

    ..  math::

        \frac{\partial \psi}{\partial \boldsymbol{C}} &=
            \sum_\alpha \frac{\partial \psi}{\partial \lambda_\alpha}
                \frac{\partial \lambda_\alpha}{\partial \boldsymbol{C}}

        \boldsymbol{S} &= 2 \frac{\partial \psi}{\partial \boldsymbol{C}}

    Furthermore, the second partial derivatives of the elasticity tensor are carried
    out.

    ..  math::

        \frac{\partial^2 \psi}{\partial \boldsymbol{C}~\partial \boldsymbol{C}} &=
            \sum_\alpha \sum_\beta \frac{\partial^2 \psi}
            {\partial \lambda_\alpha~\partial \lambda_\beta}
            \frac{\partial \lambda_\alpha}{\partial \boldsymbol{C}} \otimes
            \frac{\partial \lambda_\beta}{\partial \boldsymbol{C}}

            &+ \sum_\alpha \sum_{\beta \ne \alpha} \frac{
                \frac{\partial \psi}{\partial \lambda^2_\alpha} -
                \frac{\partial \psi}{\partial \lambda^2_\beta}
            }{\lambda^2_\alpha - \lambda^2_\beta} \left( \boldsymbol{M}_\alpha \odot
                \boldsymbol{M}_\beta + \boldsymbol{M}_\beta \odot
            \boldsymbol{M}_\alpha \right)


    ..  math::

        \mathbb{C} = 4 \frac{\partial^2 \psi}
            {\partial \boldsymbol{C}~\partial \boldsymbol{C}}

    In case of repeated equal principal stretches, the rule of d'Hospital is applied.

    ..  math::

        \lim_{\lambda^2_\beta \rightarrow \lambda^2_\alpha} \left(
                \frac{\frac{\partial \psi}{\partial \lambda^2_\alpha} -
                    \frac{\partial \psi}{\partial \lambda^2_\beta}
                }{\lambda^2_\alpha - \lambda^2_\beta}
            \right) = \left(
            - \frac{\partial^2 \psi}{\partial \lambda^2_\alpha~\partial \lambda^2_\beta}
            + \frac{\partial^2 \psi}{\partial \lambda^2_\beta~\partial \lambda^2_\beta}
        \right)

    """

    def __init__(self, material, nstatevars=0, parallel=False):
        """Initialize the Framework for an invariant-based isotropic hyperelastic
        material formulation."""

        self.parallel = parallel
        self.material = material

        self.x = [np.eye(3), np.zeros(nstatevars)]

    def gradient(self, C, statevars):
        """The gradient as the partial derivative of the strain energy function w.r.t.
        the right Cauchy-Green deformation tensor (one half of the second Piola
        Kirchhoff stress tensor)."""

        self.λ, self.M = eigh(C, fun=np.sqrt)

        self.dWdλ, statevars_new = self.material.gradient(self.λ, statevars)
        self.dWdλ = np.asarray(self.dWdλ)

        self.dWdλC = self.dWdλ / (2 * self.λ)

        dWdC = np.sum(np.expand_dims(self.dWdλC, axis=1) * self.M, axis=0)

        return dWdC, statevars_new

    def hessian(self, C, statevars):
        """The hessian as the second partial derivatives of the strain energy function
        w.r.t. the right Cauchy-Green deformation tensor (a quarter of the Lagrangian
        fourth-order elasticity tensor associated to the second Piola-Kirchhoff stress
        tensor)."""

        dWdC, statevars = self.gradient(C, statevars)
        d2Wdλdλ = self.material.hessian(self.λ, statevars)

        if not isinstance(d2Wdλdλ, np.ndarray):
            newshape = self.λ[0].shape
            _d2Wdλdλ = np.zeros((6, *self.λ[0].shape))
            for a in range(6):
                if d2Wdλdλ[a] is not None:
                    _d2Wdλdλ[a] = np.broadcast_to(d2Wdλdλ[a], newshape)
            d2Wdλdλ = _d2Wdλdλ

        λ = self.λ
        M = self.M
        dWdλC = self.dWdλC

        a = [0, 1, 2, 0, 1, 0]
        b = [0, 1, 2, 1, 2, 2]

        d2WdλC2 = d2Wdλdλ / (4 * λ[a] * λ[b])
        d2WdλC2[:3] -= dWdλC / (2 * λ**2)

        d2WdCdC = np.zeros((6, 6, *dWdλC.shape[1:]))

        f = np.zeros_like(λ[0])

        for m, (α, β) in enumerate(zip(a, b)):
            M4 = dya(M[α], M[β])
            d2WdCdC += d2WdλC2[m] * M4

            if β != α:
                v = λ[α] ** 2 - λ[β] ** 2
                mask = np.isclose(v, 0)

                f[~mask] = (dWdλC[α][~mask] - dWdλC[β][~mask]) / v[~mask]
                f[mask] = d2WdλC2[β][mask] - d2WdλC2[m][mask]

                d2WdCdC += d2WdλC2[m] * transpose(M4) + f * 2 * cdya(M[α], M[β])

        return d2WdCdC



def strain(stretch, exponent):
    "Seth-Hill strain as a function of the stretch (with first and second derivatives)."

    k = exponent
    λ = stretch

    E = lambda λ, k: (λ**k - 1) / k  # noqa: E731
    dEdλ = lambda λ, k: λ ** (k - 1)  # noqa: E731
    d2Edλdλ = lambda λ, k: (k - 1) * λ ** (k - 2)  # noqa: E731

    E0 = E(λ=1, k=2)
    dEdλ0 = dEdλ(λ=1, k=2)
    d2Edλdλ0 = d2Edλdλ(λ=1, k=2)

    return E(λ, k), dEdλ(λ, k), d2Edλdλ(λ, k), E0, dEdλ0, d2Edλdλ0


def deformation(stretch, exponent):
    """Generalized deformation as a function of the stretch (with first and second
    derivatives)."""

    k = exponent
    λ = stretch

    Ck = lambda λ, k: λ**k  # noqa: E731
    dCkdλ = lambda λ, k: k * λ ** (k - 1)  # noqa: E731
    d2Ckdλdλ = lambda λ, k: k * (k - 1) * λ ** (k - 2)  # noqa: E731

    C0 = Ck(λ=1, k=2)
    dCdλ0 = dCkdλ(λ=1, k=2)
    d2Cdλdλ0 = d2Ckdλdλ(λ=1, k=2)

    return Ck(λ, k), dCkdλ(λ, k), d2Ckdλdλ(λ, k), C0, dCdλ0, d2Cdλdλ0



import numpy as np


def asvoigt(A, mode=2):
    r"""Convert a three-dimensional tensor from full array-storage into reduced
    symmetric (Voigt-notation) vector/matrix storage.

    Parameters
    ----------
    A : np.ndarray
        A three-dimensional second- or fourth-order tensor in full array-storage.
    mode : int, optional
        The mode, 2 for second-order and 4 for fourth-order tensors (default is 2).

    Returns
    -------
    np.ndarray
        A three-dimensional second- or fourth-order tensor in reduced symmetric (Voigt)
        vector/matrix storage.

    Notes
    -----
    This is the inverse operation of :func:`astensor`.

    For a symmetric 3x3 second-order tensor :math:`C_{ij} = C_{ji}`, the upper triangle
    entries are inserted into a 6x1 vector, starting from the main diagonal, followed by
    the consecutive next upper diagonals.

    ..  math::

        \boldsymbol{C} = \begin{bmatrix}
            C_{11} & C_{12} & C_{13} \\
            C_{12} & C_{22} & C_{23} \\
            C_{13} & C_{23} & C_{33}
        \end{bmatrix} \qquad \longrightarrow \boldsymbol{C} = \begin{bmatrix}
            C_{11} & C_{22} & C_{33} & C_{12} & C_{23} & C_{13}
        \end{bmatrix}^T

    For a (at least minor) symmetric 3x3x3x3 fourth-order tensor :math:`A_{ijkl} =
    A_{jikl} = A_{ijlk} = A_{jilk}`, rearranged to 9x9, the upper triangle entries are
    inserted into a 6x6 matrix, starting from the main diagonal, followed by the
    consecutive next upper diagonals.

    ..  math::

        \begin{bmatrix}
            A_{1111} & A_{1112} & A_{1113} &
            A_{1121} & A_{1122} & A_{1123} &
            A_{1131} & A_{1132} & A_{1133} \\
            %
            A_{1211} & A_{1212} & A_{1213} &
            A_{1221} & A_{1222} & A_{1223} &
            A_{1231} & A_{1232} & A_{1233} \\
            %
            \dots & \dots & \dots & \dots & \dots & \dots & \dots & \dots & \dots \\
            A_{3111} & A_{3112} & A_{3113} &
            A_{3121} & A_{3122} & A_{3123} &
            A_{3131} & A_{3132} & A_{3133}
            %
        \end{bmatrix} \qquad

        \longrightarrow \mathbb{A} = \begin{bmatrix}
            A_{1111} & A_{1122} & A_{1133} & A_{1112} & A_{1123} & A_{1113} \\
            A_{2211} & A_{2222} & A_{2233} & A_{2212} & A_{2223} & A_{2213} \\
             \dots   &  \dots   &  \dots   &  \dots   &  \dots   &  \dots   \\
            A_{1311} & A_{1322} & A_{1333} & A_{1312} & A_{1323} & A_{1313}
        \end{bmatrix}

    Examples
    --------
    >>> from hyperelastic.math import asvoigt
    >>> import numpy as np

    >>> C = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3)
    >>> asvoigt(C, mode=2)
    array([1. , 1.1, 1.2, 1.3, 1.4, 1.5])

    >>> A = np.einsum("ij,kl", C, C)
    >>> asvoigt(A, mode=4)
    array([[1.  , 1.1 , 1.2 , 1.3 , 1.4 , 1.5 ],
           [1.1 , 1.21, 1.32, 1.43, 1.54, 1.65],
           [1.2 , 1.32, 1.44, 1.56, 1.68, 1.8 ],
           [1.3 , 1.43, 1.56, 1.69, 1.82, 1.95],
           [1.4 , 1.54, 1.68, 1.82, 1.96, 2.1 ],
           [1.5 , 1.65, 1.8 , 1.95, 2.1 , 2.25]])

    """

    if mode == 2:  # 3x3 symmetric tensor
        i = [0, 1, 2, 0, 1, 0]
        j = [0, 1, 2, 1, 2, 2]

        return A[i, j]

    elif mode == 4:  # 3x3x3x3 major-symmetric tensor
        B = np.zeros((6, 6, *A.shape[4:]))
        for i in range(3):
            for j in range(3):
                B[i, j] = A[i, i, j, j]

        B[3, 3] = A[0, 1, 0, 1]
        B[4, 4] = A[1, 2, 1, 2]
        B[5, 5] = A[0, 2, 0, 2]

        for i in range(3):
            B[i, 3] = A[i, i, 0, 1]
            B[i, 4] = A[i, i, 1, 2]
            B[i, 5] = A[i, i, 0, 2]
            B[3, i] = A[0, 1, i, i]
            B[4, i] = A[1, 2, i, i]
            B[5, i] = A[0, 2, i, i]

        B[3, 4] = A[0, 1, 1, 2]
        B[3, 5] = A[0, 1, 0, 2]

        B[4, 3] = A[1, 2, 0, 1]
        B[4, 5] = A[1, 2, 0, 2]

        B[5, 3] = A[0, 2, 0, 1]
        B[5, 4] = A[0, 2, 1, 2]

        return B


def astensor(A, mode=2):
    r"""Convert a three-dimensional tensor from symmetric (Voigt-notation) vector/matrix
    storage into full array-storage.

    Parameters
    ----------
    A : np.ndarray
        A three-dimensional second- or fourth-order tensor in reduced symmetric (Voigt)
        vector/matrix storage.
    mode : int, optional
        The mode, 2 for second-order and 4 for fourth-order tensors (default is 2).

    Returns
    -------
    np.ndarray
        A three-dimensional second- or fourth-order tensor in full array-storage.

    Notes
    -----
    This is the inverse operation of :func:`asvoigt`.

    For a symmetric 3x3 second-order tensor :math:`C_{ij} = C_{ji}`, the entries are
    re-created from a 6x1 vector.

    ..  math::

        \boldsymbol{C} = \begin{bmatrix}
            C_{11} & C_{22} & C_{33} & C_{12} & C_{23} & C_{13}
        \end{bmatrix}^T \longrightarrow
        %
        \boldsymbol{C} = \begin{bmatrix}
            C_{11} & C_{12} & C_{13} \\
            C_{12} & C_{22} & C_{23} \\
            C_{13} & C_{23} & C_{33}
        \end{bmatrix} \qquad

    For a (at least minor) symmetric 3x3x3x3 fourth-order tensor :math:`A_{ijkl} =
    A_{jikl} = A_{ijlk} = A_{jilk}`, the entries are re-created from
    a 6x6 matrix.

    ..  math::

        \mathbb{A} = \begin{bmatrix}
            A_{1111} & A_{1122} & A_{1133} & A_{1112} & A_{1123} & A_{1113} \\
            A_{2211} & A_{2222} & A_{2233} & A_{2212} & A_{2223} & A_{2213} \\
             \dots   &  \dots   &  \dots   &  \dots   &  \dots   &  \dots   \\
            A_{1311} & A_{1322} & A_{1333} & A_{1312} & A_{1323} & A_{1313}
        \end{bmatrix}

        \longrightarrow \begin{bmatrix}
            A_{1111} & A_{1112} & A_{1113} &
            A_{1121} & A_{1122} & A_{1123} &
            A_{1131} & A_{1132} & A_{1133} \\
            %
            A_{1211} & A_{1212} & A_{1213} &
            A_{1221} & A_{1222} & A_{1223} &
            A_{1231} & A_{1232} & A_{1233} \\
            %
            \dots & \dots & \dots & \dots & \dots & \dots & \dots & \dots & \dots \\
            A_{3111} & A_{3112} & A_{3113} &
            A_{3121} & A_{3122} & A_{3123} &
            A_{3131} & A_{3132} & A_{3133}
            %
        \end{bmatrix} \qquad

    Examples
    --------
    >>> from hyperelastic.math import asvoigt, astensor
    >>> import numpy as np

    >>> C = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3)
    >>> C6 = asvoigt(C, mode=2)
    >>> D = astensor(C6, mode=2)
    >>> np.allclose(C, D)
    True

    >>> D
    array([[1. , 1.3, 1.5],
           [1.3, 1.1, 1.4],
           [1.5, 1.4, 1.2]])

    >>> A = np.einsum("ij,kl", C, C)
    >>> A66 = asvoigt(A, mode=4)
    >>> B = astensor(A66, mode=4)
    >>> np.allclose(A, B)
    True

    """

    if mode == 2:  # second order tensor of shape 6 to 3x3
        a = np.array([0, 3, 5, 3, 1, 4, 5, 4, 2]).reshape(3, 3)
        return A[a]

    elif mode == 4:  # fourth order tensor of shape 6x6 to 3x3x3x3
        a = np.array([0, 3, 5, 3, 1, 4, 5, 4, 2])
        i, j = np.meshgrid(a, a, indexing="ij")
        return A[i.reshape(3, 3, 3, 3), j.reshape(3, 3, 3, 3)]


def tril_from_triu(A, dim=6, out=None):
    "Copy upper triangle values to lower triangle values of a nxn tensor inplace."

    if out is None:
        out = A.copy()

    i, j = np.triu_indices(dim, 1)
    out[j, i] = A[i, j]
    return out


def triu_from_tril(A, dim=6, out=None):
    "Copy lower triangle values to upper triangle values of a nxn tensor inplace."

    if out is None:
        out = A.copy()

    i, j = np.tril_indices(dim, -1)
    out[j, i] = A[i, j]
    return out


def trace(A):
    r"""The trace of a symmetric second-order tensor in reduced vector storage as the
    sum of the main diagonal.

    Parameters
    ----------
    A : np.ndarray
        Symmetric second-order tensor in reduced vector storage.

    Returns
    -------
    np.ndarray
        Trace of a symmetric second-order tensor in reduced vector storage.

    Notes
    -----

    ..  math::

        \text{tr}\left(\boldsymbol{C}\right) &= \boldsymbol{C} : \boldsymbol{I}
            = C_{11} + C_{22} + C_{33}

        C_{kk} &= C_{ij} : \delta_{ij}

    Examples
    --------
    >>> from hyperelastic.math import asvoigt, trace
    >>> import numpy as np

    >>> A = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3)

    >>> trA = trace(asvoigt(A))
    >>> trA
    3.3

    >>> np.allclose(trA, np.trace(A))
    True

    """

    return np.sum(A[:3], axis=0)


def transpose(A):
    r"""The major-transpose of a symmetric fourth-order tensor in reduced vector
    storage.

    Parameters
    ----------
    A : np.ndarray
        Symmetric fourth-order tensor in reduced vector storage.

    Returns
    -------
    np.ndarray
        Major-transpose of a symmetric fourth-order tensor in reduced vector storage.

    Notes
    -----

    ..  math::

        \mathbb{A}_{ijkl}^T = \mathbb{A}_{klij}


    Examples
    --------
    >>> from hyperelastic.math import asvoigt, dya, transpose
    >>> import numpy as np

    >>> A = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3)
    >>> B = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2])[::-1].reshape(3, 3)

    >>> C = dya(asvoigt(A), asvoigt(B))
    >>> C
    array([[1.2 , 1.1 , 1.  , 1.4 , 1.3 , 1.5 ],
           [1.32, 1.21, 1.1 , 1.54, 1.43, 1.65],
           [1.44, 1.32, 1.2 , 1.68, 1.56, 1.8 ],
           [1.56, 1.43, 1.3 , 1.82, 1.69, 1.95],
           [1.68, 1.54, 1.4 , 1.96, 1.82, 2.1 ],
           [1.8 , 1.65, 1.5 , 2.1 , 1.95, 2.25]])

    >>> transpose(C)
    array([[1.2 , 1.32, 1.44, 1.56, 1.68, 1.8 ],
           [1.1 , 1.21, 1.32, 1.43, 1.54, 1.65],
           [1.  , 1.1 , 1.2 , 1.3 , 1.4 , 1.5 ],
           [1.4 , 1.54, 1.68, 1.82, 1.96, 2.1 ],
           [1.3 , 1.43, 1.56, 1.69, 1.82, 1.95],
           [1.5 , 1.65, 1.8 , 1.95, 2.1 , 2.25]])

    >>> D = astensor(C, mode=4)
    >>> E = astensor(transpose(C), mode=4)
    >>> np.allclose(D, np.einsum("klij", E))
    True

    """

    return np.swapaxes(A, 0, 1)


def det(A):
    """The determinant of a symmetric second-order tensor in reduced vector storage.

    Parameters
    ----------
    A : np.ndarray
        Symmetric second-order tensor in reduced vector storage.

    Returns
    -------
    np.ndarray
        Determinant of a symmetric second-order tensor in reduced vector storage.

    Examples
    --------
    >>> from hyperelastic.math import asvoigt, det
    >>> import numpy as np

    >>> A = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3)
    >>> B = asvoigt(A)

    >>> det(B)
    0.3169999999999993

    >>> np.allclose(det(B), np.linalg.det(A))
    True

    """

    return (
        A[0] * A[1] * A[2]
        + 2 * A[3] * A[4] * A[5]
        - A[4] ** 2 * A[0]
        - A[5] ** 2 * A[1]
        - A[3] ** 2 * A[2]
    )


def inv(A, determinant=None, out=None):
    r"""The inverse of a symmetric second-order tensor in reduced vector storage.

    Parameters
    ----------
    A : np.ndarray
        Symmetric second-order tensor in reduced vector storage.
    determinant : np.ndarray or None, optional
        The determinant of the symmetric second-order tensor (default is None).
    out : np.ndarray or None, optional
        A location into which the result is stored. If provided, it must have a shape
        that the inputs broadcast to. If not provided or None, a freshly-allocated array
        is returned (default is None).

    Returns
    -------
    np.ndarray
        Inverse of a symmetric second-order tensor in reduced vector storage.

    Notes
    -----

    ..  math::

        \boldsymbol{C} \boldsymbol{C}^{-1} = \boldsymbol{I}


    Examples
    --------
    >>> from hyperelastic.math import asvoigt, inv
    >>> import numpy as np

    >>> A = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3)
    >>> B = asvoigt(A)

    >>> inv(B)
    array([-2.01892744, -3.31230284, -1.86119874,  1.70347003,  1.73501577, 0.5362776 ])

    >>> np.allclose(dot(B, inv(B)), np.eye(3))
    True

    """

    detAinvA = np.zeros_like(A)

    if determinant is None:
        detA = det(A)
    else:
        detA = determinant

    Ai = A[[1, 0, 0, 4, 3, 3]]
    Aj = A[[2, 2, 1, 5, 5, 4]]
    Ak = A[[4, 5, 3, 3, 0, 1]]
    Al = A[[4, 5, 3, 2, 4, 5]]

    if out is None:
        out = Ai

    Aij = np.multiply(Ai, Aj, out=Ai)
    Akl = np.multiply(Ak, Al, out=Ak)

    detAinvA = np.subtract(Aij, Akl, out=out)

    return np.divide(*np.broadcast_arrays(detAinvA, detA), out=out)


def dot(A, B, mode=(2, 2)):
    r"""The dot product of two symmetric second-order tensors in reduced vector storage,
    where the second index of the first tensor and the first index of the second tensor
    are contracted.

    Parameters
    ----------
    A : np.ndarray
        First symmetric second-order tensor in reduced vector storage.
    B : np.ndarray
        Second symmetric second-order tensor in reduced vector storage.
    mode : 2-tuple, optional
        The mode, 2 for second-order and 4 for fourth-order tensors (default is (2, 2)).

    Returns
    -------
    np.ndarray
        Dot product of two symmetric second-order tensors in reduced vector storage.

    Notes
    -----

    ..  math::

        C &= \boldsymbol{A} ~ \boldsymbol{B}

        C_{ij} &= A_{ik} B_{kj}

    Examples
    --------
    >>> from hyperelastic.math import asvoigt, dot
    >>> import numpy as np

    >>> A = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3)
    >>> B = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2])[::-1].reshape(3, 3)

    >>> C = dot(asvoigt(A), asvoigt(B), mode=(2, 2))
    >>> C
    array([[5.27, 4.78, 4.69],
           [5.2 , 4.85, 4.78],
           [5.56, 5.2 , 5.27]])

    >>> D = A @ B
    >>> np.allclose(C, D)
    True

    """

    if mode == (2, 2):
        trax = np.broadcast_shapes(A.shape[1:], B.shape[1:])
        C = np.zeros((3, 3, *trax))
        C[0, 0] = A[0] * B[0] + A[3] * B[3] + A[5] * B[5]
        C[1, 1] = A[3] * B[3] + A[1] * B[1] + A[4] * B[4]
        C[2, 2] = A[4] * B[4] + A[5] * B[5] + A[2] * B[2]
        C[0, 1] = A[0] * B[3] + A[3] * B[1] + A[5] * B[4]
        C[1, 2] = A[3] * B[5] + A[1] * B[4] + A[4] * B[2]
        C[0, 2] = A[0] * B[5] + A[3] * B[4] + A[5] * B[2]
        C[1, 0] = A[3] * B[0] + A[1] * B[3] + A[4] * B[5]
        C[2, 1] = A[5] * B[3] + A[4] * B[1] + A[2] * B[4]
        C[2, 0] = A[5] * B[0] + A[4] * B[3] + A[2] * B[5]

    return C


def eye(A=None):
    r"""A 3x3 tensor in Voigt-storage with ones on the diagonal and zeros elsewhere. The
    dimension is taken from the input argument (a symmetric second-order tensor in
    reduced vector storage).

    Parameters
    ----------
    A : np.ndarray or None, optional
        Symmetric second- or fourth-order tensor in reduced vector storage (default is
        None).

    Returns
    -------
    np.ndarray
        Identity matrix in reduced vector storage.

    Notes
    -----
    ..  math::

        \boldsymbol{I} = \begin{bmatrix} 1 & 1 & 1 & 0 & 0 & 0 \end{bmatrix}^T

    Examples
    --------
    >>> from hyperelastic.math import asvoigt, eye
    >>> import numpy as np

    >>> A = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3)
    >>> B = asvoigt(A)
    >>> eye(B)
    array([1., 1., 1., 0., 0., 0.])

    """

    trax = ()
    if A is not None:
        trax = np.ones(len(A.shape[1:]), dtype=int)

    return np.array([1, 1, 1, 0, 0, 0], dtype=float).reshape(6, *trax)


def ddot(A, B, mode=(2, 2)):
    r"""The double-dot product of two symmetric tensors in reduced vector storage, where
    the two innermost indices of both tensors are contracted.

    Parameters
    ----------
    A : np.ndarray
        First symmetric second- or fourth-order tensor in reduced vector storage.
    B : np.ndarray
        Second symmetric second- or fourth-order tensor in reduced vector storage.
    mode : 2-tuple, optional
        The mode, 2 for second-order and 4 for fourth-order tensors (default is (2, 2)).

    Returns
    -------
    np.ndarray
        Double-dot product of two symmetric tensors in scalar or reduced vector/matrix
        storage.

    Notes
    -----

    ..  math::

        C &= \boldsymbol{A} : \boldsymbol{B}

        C &= A_{ij} : B_{ij}


    ..  math::

        \boldsymbol{C} &= \boldsymbol{A} : \mathbb{B}

        C_{kl} &= A_{ij} : \mathbb{B}_{ijkl}


    ..  math::

        \boldsymbol{C} &= \mathbb{B} : \boldsymbol{A}

        C_{ij} &= \mathbb{B}_{ijkl} : A_{kl}


    ..  math::

        \mathbb{C} &= \mathbb{A} : \mathbb{B}

        \mathbb{C}_{ijmn} &= \mathbb{A}_{ijkl} : \mathbb{A}_{klmn}


    Examples
    --------
    >>> from hyperelastic.math import asvoigt, ddot
    >>> import numpy as np

    >>> A = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3)
    >>> B = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2])[::-1].reshape(3, 3)

    >>> A4 = np.einsum("ij,kl", A, B)
    >>> B4 = (np.einsum("ik,jl", A, A) + np.einsum("il,kj", A, A)) / 2

    >>> ddot(asvoigt(A), asvoigt(B), mode=(2, 2))
    15.39

    >>> ddot(asvoigt(A), asvoigt(B4, mode=4), mode=(2, 4))
    array([18.899, 18.863, 21.698, 18.903, 20.253, 20.316])

    >>> ddot(asvoigt(B4, mode=4), asvoigt(A), mode=(4, 2))
    array([18.899, 18.863, 21.698, 18.903, 20.253, 20.316])

    >>> ddot(asvoigt(A4, mode=4), asvoigt(B4, mode=4), mode=(4, 4))
    array([[18.519 , 18.787 , 21.944 , 18.675 , 20.326 , 20.225 ],
           [20.3709, 20.6657, 24.1384, 20.5425, 22.3586, 22.2475],
           [22.2228, 22.5444, 26.3328, 22.41  , 24.3912, 24.27  ],
           [24.0747, 24.4231, 28.5272, 24.2775, 26.4238, 26.2925],
           [25.9266, 26.3018, 30.7216, 26.145 , 28.4564, 28.315 ],
           [27.7785, 28.1805, 32.916 , 28.0125, 30.489 , 30.3375]])

    """

    weights = np.array([1, 1, 1, 2, 2, 2])
    if mode == (2, 2):
        return np.einsum("i...,i...,i->...", A, B, weights)
    elif mode == (4, 4):
        return np.einsum("ik...,kj...,k->ij...", A, B, weights)
    elif mode == (4, 2):
        return np.einsum("ij...,j...,j->i...", A, B, weights)
    elif mode == (2, 4):
        return np.einsum("i...,ij...,i->j...", A, B, weights)


def dya(A, B, out=None):
    r"""The dyadic product of two symmetric second-order tensors in reduced vector
    storage.

    Parameters
    ----------
    A : np.ndarray
        First symmetric second-order tensor in reduced vector storage.
    B : np.ndarray
        Second symmetric second-order tensor in reduced vector storage.
    out : np.ndarray or None, optional
        A location into which the result is stored. If provided, it must have a shape
        that the inputs broadcast to. If not provided or None, a freshly-allocated array
        is returned (default is None).

    Returns
    -------
    np.ndarray
        Dyadic product in reduced matrix storage.

    Notes
    -----
    The result of the dyadic product of two symmetric second order tensors
    is a minor- (but not major-) symmetric fourth-order tensor. For the case of
    :math:`\boldsymbol{A} = \boldsymbol{B}`, the result is both major- and minor-
    symmetric.

    ..  math::

        \mathbb{C} &= \boldsymbol{A} \otimes \boldsymbol{B}

        \mathbb{C}_{ijkl} &= A_{ij}~B_{kl}

    Examples
    --------
    >>> from hyperelastic.math import asvoigt, astensor, dya
    >>> import numpy as np

    >>> C = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3)
    >>> D = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2])[::-1].reshape(3, 3)

    >>> A = np.einsum("ij,kl", C, D)

    >>> C6 = asvoigt(C, mode=2)
    >>> D6 = asvoigt(D, mode=2)

    >>> np.allclose(A, astensor(dya(C6, D6), mode=4))
    True

    """

    return np.multiply(
        A.reshape(-1, 1, *A.shape[1:]),
        B.reshape(1, -1, *B.shape[1:]),
        out=out,
    )


def dev(A):
    r"""The deviatoric part of a three-dimensional second-order tensor.

    Parameters
    ----------
    A : np.ndarray
        Symmetric second-order tensor in reduced vector storage.

    Returns
    -------
    np.ndarray
        Deviatoric part of the symmetric second-order tensor in reduced vector storage.

    Notes
    -----
    ..  math::

        \text{dev}(\boldsymbol{C}) =
         \boldsymbol{C} - \frac{\text{tr}(\boldsymbol{C})}{3} \boldsymbol{I}

    """
    return A - trace(A) / 3 * eye(A)


def cdya_ik(A, B, out=None):
    r"""The overlined-dyadic product of two symmetric second-order
    tensors in reduced vector storage, where the inner indices (the second index of the
    first tensor and the first index of the second tensor) are interchanged.

    Parameters
    ----------
    A : np.ndarray
        First symmetric second-order tensor in reduced vector storage.
    B : np.ndarray
        Second symmetric second-order tensor in reduced vector storage.
    out : np.ndarray or None, optional
        A location into which the result is stored. If provided, it must have a shape
        that the inputs broadcast to. If not provided or None, a freshly-allocated array
        is returned (default is None).

    Returns
    -------
    np.ndarray
        Overlined-dyadic product in full-array storage.

    Notes
    -----
    The result of the overlined-dyadic product of two symmetric second order tensors
    is a major- (but not minor-) symmetric fourth-order tensor. This is also the case
    for :math:`\boldsymbol{A} = \boldsymbol{B}`.

    ..  math::

        \mathbb{C} &= \boldsymbol{A} \overline{\otimes} \boldsymbol{B}

        \mathbb{C}_{ijkl} &= A_{ik}~B_{jl}

    Examples
    --------
    >>> from hyperelastic.math import asvoigt, cdya_ik
    >>> import numpy as np

    >>> C = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3)
    >>> D = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2])[::-1].reshape(3, 3)

    >>> A = np.einsum("ik,jl", C, D)

    >>> C6 = asvoigt(C, mode=2)
    >>> D6 = asvoigt(D, mode=2)

    >>> np.allclose(A, cdya_ik(C6, D6))
    True

    """

    return np.swapaxes(astensor(dya(A, B, out=out), mode=4), 1, 2)


def cdya_il(A, B, out=None):
    r"""The underlined-dyadic product of two symmetric second-order
    tensors in reduced vector storage, where the right indices of the two tensors are
    interchanged.

    Parameters
    ----------
    A : np.ndarray
        First symmetric second-order tensor in reduced vector storage.
    B : np.ndarray
        Second symmetric second-order tensor in reduced vector storage.
    out : np.ndarray or None, optional
        A location into which the result is stored. If provided, it must have a shape
        that the inputs broadcast to. If not provided or None, a freshly-allocated array
        is returned (default is None).

    Returns
    -------
    np.ndarray
        Underlined-dyadic product in full-array storage.

    Notes
    -----
    The result of the underlined-dyadic product of two symmetric second
    order tensors is a non-symmetric fourth-order tensor. In case of
    :math:`\boldsymbol{A} = \boldsymbol{B}`, the fourth-order tensor is major-symmetric.

    ..  math::

        \mathbb{C} &= \boldsymbol{A} \underline{\otimes} \boldsymbol{B}

        \mathbb{C}_{ijkl} &= A_{il}~B_{kj}

    Examples
    --------
    >>> from hyperelastic.math import asvoigt, cdya_il
    >>> import numpy as np

    >>> C = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3)
    >>> D = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2])[::-1].reshape(3, 3)

    >>> A = np.einsum("il,kj", C, D)

    >>> C6 = asvoigt(C, mode=2)
    >>> D6 = asvoigt(D, mode=2)

    >>> np.allclose(A, cdya_il(C6, D6))
    True

    """

    return np.swapaxes(astensor(dya(A, B, out=out), mode=4), 1, 3)


def cdya(A, B, out=None):
    r"""The full-symmetric crossed-dyadic product of two symmetric second-order tensors
    in reduced vector storage.

    Parameters
    ----------
    A : np.ndarray
        First symmetric second-order tensor in reduced vector storage.
    B : np.ndarray
        Second symmetric second-order tensor in reduced vector storage.
    out : np.ndarray or None, optional
        A location into which the result is stored. If provided, it must have a shape
        that the inputs broadcast to. If not provided or None, a freshly-allocated array
        is returned (default is None).

    Returns
    -------
    np.ndarray
        Symmetric crossed-dyadic product in reduced matrix storage.

    Notes
    -----
    The result of the symmetric crossed-dyadic product of two symmetric second order
    tensors is a minor- and major-symmetric fourth-order tensor.

    ..  math::

        \mathbb{C} &= \frac{1}{2} \left(
            \boldsymbol{A} \odot \boldsymbol{B} + \boldsymbol{B} \odot \boldsymbol{A}
        \right)

        \mathbb{C} &= \frac{1}{4} \left(
            \boldsymbol{A} \overline{\otimes} \boldsymbol{B} +
            \boldsymbol{A} \underline{\otimes} \boldsymbol{B} +
            \boldsymbol{B} \overline{\otimes} \boldsymbol{A} +
            \boldsymbol{B} \underline{\otimes} \boldsymbol{A}
        \right)

        \mathbb{C}_{ijkl} &= \frac{1}{4} \left(
            A_{ik}~B_{jl} + A_{il}~B_{kj} + B_{ik}~A_{jl} + B_{il}~A_{kj}
        \right)

    ..  note::
        The technical implementation is based on an answer of Jérôme Richard (see
        `stackoverflow <https://stackoverflow.com/questions/76640596>`_).

    Examples
    --------
    >>> from hyperelastic.math import asvoigt, astensor, cdya
    >>> import numpy as np

    >>> C = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2]).reshape(3, 3)
    >>> D = np.array([1.0, 1.3, 1.5, 1.3, 1.1, 1.4, 1.5, 1.4, 1.2])[::-1].reshape(3, 3)

    >>> CD = (np.einsum("ik,jl", C, D) + np.einsum("il,kj", C, D)) / 2
    >>> DC = (np.einsum("ik,jl", D, C) + np.einsum("il,kj", D, C)) / 2

    >>> A = (CD + DC) / 2

    >>> C6 = asvoigt(C, mode=2)
    >>> D6 = asvoigt(D, mode=2)
    >>> cdya(C6, D6)
    array([[1.2   , 1.82  , 2.25  , 1.48  , 2.025 , 1.65  ],
           [1.82  , 1.21  , 1.82  , 1.485 , 1.485 , 1.825 ],
           [2.25  , 1.82  , 1.2   , 2.025 , 1.48  , 1.65  ],
           [1.48  , 1.485 , 2.025 , 1.515 , 1.7375, 1.7575],
           [2.025 , 1.485 , 1.48  , 1.7375, 1.515 , 1.7575],
           [1.65  , 1.825 , 1.65  , 1.7575, 1.7575, 1.735 ]])

    >>> np.allclose(A, astensor(cdya(C6, D6), mode=4))
    True

    >>> np.allclose(A, astensor(cdya(D6, C6), mode=4))
    True

    """

    i, j = [a.ravel() for a in np.triu_indices(6)]

    a = np.array([(0, 0), (1, 1), (2, 2), (0, 1), (1, 2), (0, 2)])
    b = np.array([0, 3, 5, 3, 1, 4, 5, 4, 2]).reshape(3, 3)

    i, j, k, l = np.hstack([a[i], a[j]]).T

    ij = b[i, j]
    kl = b[k, l]

    ik = b[i, k]
    jl = b[j, l]

    il = b[i, l]
    kj = b[k, j]

    if out is None:
        out = np.zeros((6, *np.broadcast_shapes(A.shape, B.shape)))

    A, B = np.broadcast_arrays(A, B)
    Aik, Bjl, Ail, Bkj = A[ik], B[jl], A[il], B[kj]

    values = np.multiply(Aik, Bjl, out=Aik) + np.multiply(Ail, Bkj, out=Ail)

    if A is not B:
        Bik, Ajl, Bil, Akj = B[ik], A[jl], B[il], A[kj]

        values += np.multiply(Bik, Ajl, out=Bik) + np.multiply(Bil, Akj, out=Bil)
        values /= 4
    else:
        values /= 2

    out[ij, kl] = out[kl, ij] = values

    return out


def eigh(A, fun=None):
    "Eigenvalues and -bases of matrix A."
    wA, vA = np.linalg.eigh(astensor(A).T)
    MA = np.einsum("...ia,...ja->...ija", vA, vA)
    if fun is not None:
        wA = fun(wA)
    return wA.T, np.array([asvoigt(M) for M in MA.T])



import numpy as np

import hyperelastic


def pre(diameter, length):
    "Generate a synthetic experiment."
    area = diameter**2 * np.pi / 4
    stretch = np.linspace(1, 5)

    force = stretch - 1 / stretch**2 + (stretch - 1) ** 4 / 50

    stretch2 = stretch[::-1]
    stretch2 = (stretch2 - stretch2.max()) * 0.9 + stretch2.max()
    stretch = np.concatenate([stretch, stretch2])

    force2 = force * 0.5
    flow = np.linspace(1, 0, 15) ** 0.5
    force2[-15:] = flow * force2[-15:] + (1 - flow) * force[-15:]

    force = np.concatenate([force, force2[::-1]]) * area

    np.random.seed(56489)
    force += (stretch - 1) * (np.random.rand(len(stretch)) - 0.5) * 30

    displacement = (stretch - 1) * length

    return force, displacement, area, length


def material(**kwargs):
    k = kwargs.pop("k")
    tod = hyperelastic.models.invariants.ThirdOrderDeformation(strain=False, **kwargs)
    fun = hyperelastic.models.generalized.deformation
    framework = hyperelastic.GeneralizedInvariantsFramework(tod, fun=fun, exponent=k)
    # framework = hyperelastic.InvariantsFramework(tod)
    return hyperelastic.DistortionalSpace(framework)


def test_lab():
    force, displacement, area, length = pre(diameter=15, length=100)

    experiments = [
        hyperelastic.lab.Experiment(
            label="Uniaxial Tension",
            displacement=displacement,
            force=force,
            area=area,
            length=length,
        ),
        hyperelastic.lab.Experiment(
            label="Planar Tension",
            displacement=displacement[::2],
            force=force[::2] * 8 / 7,
            area=area,
            length=length,
        ),
        hyperelastic.lab.Experiment(
            label="Biaxial Tension",
            displacement=displacement[::2] / 2,
            force=force[::2] * 2 / 3,
            area=area,
            length=length,
        ),
    ]

    ux = hyperelastic.lab.Uniaxial()
    bx = hyperelastic.lab.Biaxial()
    ps = hyperelastic.lab.Planar()

    simulations = [
        hyperelastic.lab.Simulation(
            ux,
            experiments[0].stretch,
            material=material,
            labels=["C10", "C20", "C30", "k"],
        ),
        hyperelastic.lab.Simulation(
            ps,
            experiments[1].stretch,
            material=material,
            labels=["C10", "C20", "C30", "k"],
        ),
        hyperelastic.lab.Simulation(
            bx,
            experiments[2].stretch,
            material=material,
            labels=["C10", "C20", "C30", "k"],
        ),
    ]

    optimize = hyperelastic.lab.Optimize(
        experiments=experiments,
        simulations=simulations,
        parameters=np.ones(4),
        mask=[  # consider only uploading path
            np.diff(experiments[0].displacement, prepend=0) >= 0,
            np.diff(experiments[1].displacement, prepend=0) >= 0,
            np.diff(experiments[2].displacement, prepend=0) >= 0,
            # np.zeros_like(experiments[1].displacement, dtype=bool), # deactivate
        ],
    )

    parameters, pcov = optimize.curve_fit(method="lm")

    assert np.allclose(
        parameters, [5.22901342e-01, 6.60281220e-03, 1.21021751e-03, 1.54827834e00]
    )

    fig, ax = experiments[0].plot_force_displacement()
    fig, ax = experiments[1].plot_force_displacement(ax=ax)
    fig, ax = experiments[2].plot_force_displacement(ax=ax)
    ax.legend()

    fig, ax = experiments[0].plot_stress_stretch()
    fig, ax = experiments[1].plot_stress_stretch(ax=ax)
    fig, ax = experiments[2].plot_stress_stretch(ax=ax)
    ax.legend()

    fig, ax = simulations[0].plot_stress_stretch()
    fig, ax = simulations[1].plot_stress_stretch(ax=ax)
    fig, ax = simulations[2].plot_stress_stretch(ax=ax)
    ax.legend()
    ax.set_title("Yeoh (Generalized Invariants Framework)")

    fig, ax = optimize.plot(title="Yeoh (Generalized Invariants Framework)")
    ax.set_xlim(None, 1.1 * ax.get_xlim()[1])

    return parameters


if __name__ == "__main__":
    parameters = test_lab()
    print(parameters)



class ThirdOrderDeformation:
    r"""Third Order Deformation isotropic hyperelastic material formulation based on the
    first and second invariant of the right Cauchy-Green deformation tensor. The strain
    energy density per unit undeformed volume is given as a sum of multivariate
    polynomials.

    ..  math::

        \psi(I_1, I_2) =
            \sum_{(i+j) \ge 1}^{(i+2j) \le 3} C_{ij}~(I_1 - 3)^i (I_2 - 3)^j

    The first partial derivatives of the strain energy density w.r.t. the
    invariants are given below.

    ..  math::

        \frac{\partial \psi}{\partial I_1} &=
            \sum_{i \ge 1} C_{ij}~i~(I_1 - 3)^{i-1} (I_2 - 3)^j

        \frac{\partial \psi}{\partial I_2} &=
            \sum_{j \ge 1} C_{ij}~(I_1 - 3)^i~j~(I_2 - 3)^{j-1}

    Furthermore, the second partial derivatives of the strain energy density w.r.t. the
    invariants are carried out.

    ..  math::

        \frac{\partial^2 \psi}{\partial I_1~\partial I_1} &=
            \sum_{i \ge 2} C_{ij}~i~(i-1)~(I_1 - 3)^{i-2} (I_2 - 3)^j

        \frac{\partial^2 \psi}{\partial I_2~\partial I_2} &=
            \sum_{j \ge 2} C_{ij}~(I_1 - 3)^i~j~(j-1)~(I_2 - 3)^{j-2}

        \frac{\partial^2 \psi}{\partial I_1~\partial I_2} &=
            \sum_{i \ge 1, j \ge 1} C_{ij}~i~(I_1 - 3)^{i-1}~j~(I_2 - 3)^{j-1}


    """

    def __init__(self, C10=0, C01=0, C11=0, C20=0, C30=0, strain=False):
        """Initialize the Third Order Deformation material formulation with its
        parameters.
        """

        self.C10 = C10
        self.C01 = C01
        self.C11 = C11
        self.C20 = C20
        self.C30 = C30

        # value of invariants at the undeformed state
        self.undeformed = 0 if strain else 3

    def gradient(self, I1, I2, I3, statevars):
        """The gradient as the partial derivative of the strain energy function w.r.t.
        the invariants."""

        dWdI1 = None
        dWdI2 = None
        dWdI3 = None

        if self.C10 != 0 or self.C11 != 0 or self.C20 != 0 or self.C30 != 0:
            dWdI1 = 0

        if self.C01 != 0 or self.C11 != 0:
            dWdI2 = 0

        if self.C10 != 0:
            dWdI1 += self.C10

        if self.C01 != 0:
            dWdI2 += self.C01

        if self.C11 != 0:
            dWdI1 += self.C11 * (I2 - self.undeformed)
            dWdI2 += self.C11 * (I1 - self.undeformed)

        if self.C20 != 0:
            dWdI1 += self.C20 * 2 * (I1 - self.undeformed)

        if self.C30 != 0:
            dWdI1 += self.C30 * 3 * (I1 - self.undeformed) ** 2

        return dWdI1, dWdI2, dWdI3, statevars

    def hessian(self, I1, I2, I3, statevars):
        """The hessian as the second partial derivatives of the strain energy function
        w.r.t. the invariants."""

        d2WdI1I1 = None
        d2WdI2I2 = None
        d2WdI3I3 = None
        d2WdI1I2 = None
        d2WdI2I3 = None
        d2WdI1I3 = None

        if self.C20 != 0 or self.C30 != 0:
            d2WdI1I1 = self.C20 * 2 + self.C30 * 6 * (I1 - self.undeformed)

        if self.C11 != 0:
            d2WdI1I2 = self.C11

        return d2WdI1I1, d2WdI2I2, d2WdI3I3, d2WdI1I2, d2WdI2I3, d2WdI1I3



import numpy as np

import hyperelastic
from hyperelastic import lab


def test_simulation():
    stretch = np.linspace(0.7, 2.5, 181)

    def material(**kwargs):
        "A third-order deformation material formulation."

        tod = hyperelastic.models.invariants.ThirdOrderDeformation(**kwargs)
        framework = hyperelastic.InvariantsFramework(tod)

        return hyperelastic.DeformationSpace(framework)

    simulation = lab.Simulation(
        loadcase=lab.Uniaxial(),
        stretch=stretch,
        material=material,
        labels=["C10", "C01", "C11", "C20", "C30"],
        parameters=[0.4, 0.1, 0.02, -0.04, 0.01],
    )

    fig, ax = simulation.plot_stress_stretch()
    ax.legend()
    ax.set_title("Third-Order Deformation")

    return fig


if __name__ == "__main__":
    fig = test_simulation()
    fig.savefig("../docs/hyperelastic/images/fig_simulation-tod.png")



import numpy as np


class IncompressibleHomogeneousStretch:
    r"""An incompressible homogeneous stretch load case with a longitudinal stretch and
    perpendicular transverse stretches in principal directions. This class is intended
    to be subclassed by another class with a `.defgrad()` method for the evaluation of
    the deformation gradient as utilized by the :class:`Uniaxial <.lab.Uniaxial>`,
    :class:`Planar <.lab.Planar>` and :class:`Biaxial <.lab.Biaxial>` load cases.

    Notes
    -----
    The Cauchy stress for an incompressible material is given by

    ..  math::
        \boldsymbol{\sigma} = \boldsymbol{\sigma}' + p \boldsymbol{I}

    where the Cauchy stress is converted to the first Piola-Kirchhoff stress tensor.

    ..  math::
        \boldsymbol{P} = J \boldsymbol{\sigma} \boldsymbol{F}^{-T}

    The deformation gradient and its determinant are evaluated for the
    homogeneous incompressible deformation.

    ..  math::
        \boldsymbol{F} &= \text{diag} \left(\begin{bmatrix}
            \lambda_1 & \lambda_2 & \lambda_3 \end{bmatrix} \right)

        J &= \lambda_1 \lambda_2 \lambda_3 = 1

    This enables the evaluation of the normal force per undeformed area, where
    quantities in the traction-free transverse direction are denoted with a subscript
    :math:`(\bullet)_t`.

    ..  math::
        \frac{N}{A} = P - P_t \frac{\lambda_t}{\lambda}

    """

    def stress(self, F, P, axis=0, traction_free=-1):
        r"""Normal force per undeformed area for a given deformation gradient of an
        incompressible deformation and the first Piola-Kirchhoff stress tensor.

        Parameters
        ----------
        F : ndarray
            The deformation gradient.
        P : ndarray
            The first Piola-Kirchhoff stress tensor.
        axis : int, optional
            The primary axis where the longitudinal stretch is applied on (default is
            0).
        traction_free : int, optional
            The secondary axis where the traction-free transverse stretch results from
            the constraint of incompressibility (default is -1).

        Returns
        -------
        ndarray
            The one-dimensional normal force per undeformed area.

        """

        i = axis
        j = traction_free

        return P[i, i] - P[j, j] * F[j, j] / F[i, i]


class Uniaxial(IncompressibleHomogeneousStretch):
    r"""Incompressible uniaxial tension/compression load case.

    ..  math::
        \boldsymbol{F} = \text{diag} \left(\begin{bmatrix}
            \lambda & \frac{1}{\sqrt{\lambda}} & \frac{1}{\sqrt{\lambda}} \end{bmatrix}
            \right)

    """

    def __init__(self, label=None):
        if label is None:
            label = "Uniaxial Tension"
        self.label = label

    def defgrad(self, stretch):
        "Return the Deformation Gradient tensor from given stretches."

        x = stretch
        y = 1 / np.sqrt(stretch)
        z = np.zeros_like(stretch)

        return np.array([[x, z, z], [z, y, z], [z, z, y]])


class Biaxial(IncompressibleHomogeneousStretch):
    r"""Incompressible biaxial tension/compression load case.

    ..  math::
        \boldsymbol{F} = \text{diag} \left(\begin{bmatrix}
            \lambda & \lambda & \frac{1}{\lambda^2} \end{bmatrix}
            \right)

    """

    def __init__(self, label=None):
        if label is None:
            label = "Biaxial Tension"
        self.label = label

    def defgrad(self, stretch):
        "Return the Deformation Gradient tensor from given stretches."

        x = stretch
        y = 1 / stretch**2
        z = np.zeros_like(stretch)

        return np.array([[x, z, z], [z, x, z], [z, z, y]])


class Planar(IncompressibleHomogeneousStretch):
    r"""Incompressible planar (shear) tension/compression load case.

    ..  math::
        \boldsymbol{F} = \text{diag} \left(\begin{bmatrix}
            \lambda & 1 & \frac{1}{\lambda} \end{bmatrix}
            \right)

    """

    def __init__(self, label=None):
        if label is None:
            label = "Planar Tension"
        self.label = label

    def defgrad(self, stretch):
        "Return the Deformation Gradient tensor from given stretches."

        x = stretch
        y = 1 / stretch
        o = np.ones_like(stretch)
        z = np.zeros_like(stretch)

        return np.array([[x, z, z], [z, o, z], [z, z, y]])



import matplotlib.pyplot as plt
import numpy as np

from ._curve_fit import concatenate_curve_fit


class Optimize:
    r"""Take lists of experiments and simulations and find material parameters
    for the simulation model to obtain a best possible representation of the experiments
    by the simulations.

    Attributes
    ----------
    experiments : list
        A list of :class:`Experiment <.lab.Experiment>`.
    simulations : list
        A list of :class:`Simulation <.lab.Simulation>`.
    parameters : array_like
        The material parameters.
    mask : list of array_like
        A list of masks to take the optimization-relevant data points.

    Examples
    --------
    Three different test specimens are subjected to displacement-controlled uniaxial,
    planar and biaxial tension. The applied displacement and reaction force data is
    used to create the :class:`Experiments <.lab.Experiment>`. The test specimen for the
    uniaxial load case has a cross-sectional area of :math:`A=25~\text{mm}^2` and a
    length of :math:`L=100~\text{mm}`.

    >>> area = 25
    >>> length = 100

    Some synthetic experimental data is generated to demonstrate the capabilities of the
    optimization.

    >>> import numpy as np

    >>> displacement = np.linspace(0, 2 * length, 100)
    >>> stretch = 1 + displacement / length
    >>> force = (stretch - 1 / stretch ** 2 + (stretch - 1)**5 / 10) * area

    With this reference experimental data at hand, the list of experiments is created.
    In this example, the displacement and force data as well as the cross-sectional area
    are scaled from the synthetic uniaxial experimental data to the other (synthetic)
    experiments.

    >>> from hyperelastic import lab

    >>> experiments = [
    >>>     lab.Experiment(
    >>>         label="Uniaxial Tension",
    >>>         displacement=displacement,
    >>>         force=force,
    >>>         area=area,
    >>>         length=length,
    >>>     ),
    >>>     lab.Experiment(
    >>>         label="Planar Tension",
    >>>         displacement=displacement[::2],
    >>>         force=force[::2],
    >>>         area=area / (8 / 7),
    >>>         length=length,
    >>>     ),
    >>>     lab.Experiment(
    >>>         label="Biaxial Tension",
    >>>         displacement=displacement[::2] / 2,
    >>>         force=force[::2],
    >>>         area=area / (4 / 5),
    >>>         length=length,
    >>>     ),
    >>> ]

    A function which takes the material parameters and returns the hyperelastic
    constitutive material formulation has to be provided for the simulation objects.
    Here, we use an isotropic invariant-based
    :class:`third-order deformation <.models.invariants.ThirdOrderDeformation>`
    material formulation.

    >>> def material(**kwargs):
    >>>     "A third-order deformation material formulation."
    >>>
    >>>     tod = hyperelastic.models.invariants.ThirdOrderDeformation(**kwargs)
    >>>     framework = hyperelastic.InvariantsFramework(tod)
    >>>
    >>>     return hyperelastic.DeformationSpace(framework)

    The list of labels of the material parameters is used for all simulation objects.

    >>> labels = ["C10", "C01", "C11", "C20", "C30"]

    Next, the simulations for all three loadcases are created. It is important to take
    the stretches from the according experiments.

    >>> simulations = [
    >>>     lab.Simulation(
    >>>         loadcase=lab.Uniaxial(),
    >>>         stretch=experiments[0].stretch,
    >>>         material=material,
    >>>         labels=labels,
    >>>     ),
    >>>     lab.Simulation(
    >>>         loadcase=lab.Planar(),
    >>>         stretch=experiments[1].stretch,
    >>>         material=material,
    >>>         labels=labels,
    >>>     ),
    >>>     lab.Simulation(
    >>>         loadcase=lab.Biaxial(),
    >>>         stretch=experiments[2].stretch,
    >>>         material=material,
    >>>         labels=labels,
    >>>     ),
    >>> ]

    Both the list of experiments and the list of simulations are passed to
    :class:`Optimize <.lab.Optimize>`, where its curve-fit method acts as a
    simple wrapper for :class:`scipy.optimize.curve_fit`. The initial material
    parameters are all set to one.

    >>> optimize = lab.Optimize(
    >>>     experiments=experiments,
    >>>     simulations=simulations,
    >>>     parameters=np.ones(5),
    >>> )

    >>> parameters, pcov = optimize.curve_fit(method="lm")
    >>> parameters
    array([ 0.50430357, -0.01413309,  0.0141219 , -0.01641752,  0.00492179])

    >>> fig, ax = optimize.plot(title="Third-Order Deformation")

    ..  image:: images/fig_optimize-tod.png


    """

    def __init__(self, experiments, simulations, parameters, mask=None):
        """Take lists of experiments and simulations and find material parameters
        for the simulation model to obtain a best possible representation of the
        experiments by the simulations.

        Parameters
        ----------
        experiments : list of experiments
            A list of :class:`hyperelastic.lab.Experiment`.
        simulations : list of :class:`hyperelastic.lab.Simulations`
            A list of :class:`hyperelastic.lab.Simulations`.
        parameters : array_like
            The material parameters.
        mask : list of array_like or None, optional
            A list of masks to take the optimization-relevant data points (default is
            None).

        """

        self.experiments = experiments
        self.simulations = simulations
        self.parameters = parameters
        self.mask = mask

        if self.mask is None:
            self.mask = [slice(None)] * len(self.experiments)

    def init_curve_fit(self, *args, **kwargs):
        self.f = [simulation.stress_curve_fit for simulation in self.simulations]
        self.x = [
            experiment.stretch[mask]
            for mask, experiment in zip(self.mask, self.experiments)
        ]
        self.y = [
            experiment.stress()[mask]
            for mask, experiment in zip(self.mask, self.experiments)
        ]

        f0 = np.concatenate(
            [fi(xi, *self.parameters) for fi, xi in zip(self.f, self.x)]
        )

        self.errors = np.ones_like(self.parameters) * np.nan
        self.residuals = f0 - np.concatenate(self.y)

        for simulation in self.simulations:
            simulation.parameters = self.parameters

        return self.parameters

    def mean_relative_std(self):
        """Return the relative mean of the standard deviations of the material
        parameters, normalized by the absolute mean-values of the parameters.
        """

        return np.mean(self.errors / np.abs(self.parameters)) * 100

    def norm_residuals(self):
        "Return the norm of the residuals."

        return np.linalg.norm(self.residuals)

    def curve_fit(self, *args, **kwargs):
        "Use non-linear least squares to fit a list of functions to a list of data."

        p0 = self.init_curve_fit(*args, **kwargs)

        popt, pcov = concatenate_curve_fit(self.f, self.x, self.y, p0, *args, **kwargs)
        fopt = np.concatenate([fi(xi, *popt) for fi, xi in zip(self.f, self.x)])

        self.parameters[:] = popt
        self.errors[:] = np.sqrt(np.diag(pcov))
        self.residuals[:] = fopt - np.concatenate(self.y)

        for simulation in self.simulations:
            simulation.parameters[:] = self.parameters

        return self.parameters, pcov

    def plot(self, title=None):
        fig, ax = plt.subplots()

        for i, (simulation, experiment) in enumerate(
            zip(self.simulations, self.experiments)
        ):
            label = experiment.label
            fig, ax = simulation.plot_stress_stretch(
                f"C{i}", label=f"{label}", lw=3, ax=ax
            )
            fig, ax = experiment.plot_stress_stretch(
                f"C{i}", label=f"{label} (Experiment)", lw=0.7, ax=ax
            )

        ax.legend(loc=2)
        text = "\n".join(
            [
                rf"{n} = {p:1.2g} $\pm$ {2*s:1.2g}"
                for n, p, s in zip(simulation.labels, self.parameters, self.errors)
            ]
        )
        ax.text(
            0.98,
            0.02,
            text
            + "\n\n"
            + f"Absolute Norm of Residuals = {self.norm_residuals(): 1.4g}"
            + "\n"
            + f"Mean Std. Deviation of Parameters = {self.mean_relative_std(): 1.1f}%",
            horizontalalignment="right",
            verticalalignment="bottom",
            fontsize="small",
            transform=ax.transAxes,
        )

        if title is not None:
            ax.set_title(title)

        fig.tight_layout()
        return fig, ax



from ._plotting import LabPlotter


class Experiment(LabPlotter):
    """Results of an experiment along with methods to convert and plot the data.

    Attributes
    ----------
    label : str
        The title of the experiment.
    displacement : array_like
        The measured or applied displacement data.
    force : array_like
        The measured or applied force data.
    area : float
        The undeformed reference cross-sectional area used to evaluate the stress.
    length : float
        The undeformed reference length used to evaluate the stretch.
    time : array_like or None
        The timetrack of the measurement.
    temperature : array_like or None
        The measured or applied temperature data.
    stretch : array_like
        The stretch as the calculated ratio of the deformed vs. the undeformed length.

    """

    def __init__(
        self,
        label,
        displacement,
        force,
        area=1.0,
        length=1.0,
        time=None,
        temperature=None,
    ):
        """Initialize the results of an experiment.

        Parameters
        ----------
        label : str
            The title of the experiment.
        displacement : array_like
            The measured or applied displacement data.
        force : array_like
            The measured or applied force data.
        area : float, optional
            The undeformed reference cross-sectional area used to evaluate the stress
            (default is 1.0).
        length : float, optional
            The undeformed reference length used to evaluate the stretch (default is
            1.0).
        time : array_like or None, optional
            The timetrack of the measurement (default is None).
        temperature : array_like or None, optional
            The measured or applied temperature data (default is None).

        """

        super().__init__()

        self.label = label
        self.displacement = displacement
        self.force = force
        self.area = area
        self.length = length
        self.time = time
        self.temperature = temperature

        self.stretch = 1 + self.displacement / self.length

    def stress(self):
        "Evaluate the stress as force per undeformed area."

        return self.force / self.area



import matplotlib.pyplot as plt
import numpy as np

import hyperelastic


def test_generalized():
    ϵ = np.sqrt(np.finfo(np.float64).eps)
    λ = np.append(1, np.linspace(1, 2, 21) + ϵ)
    z = np.zeros_like(λ)

    # uniaxial incompressible deformation
    defgrad = np.array([[λ, z, z], [z, 1 / np.sqrt(λ), z], [z, z, 1 / np.sqrt(λ)]])

    # loop over strain exponents
    for k in [-0.5, 0.001, 2.0, 4.2]:
        tod = hyperelastic.models.invariants.ThirdOrderDeformation(
            strain=False, C10=0.3, C01=0.2, C20=-0.1, C30=0.05, C11=0.1
        )
        fun = hyperelastic.models.generalized.deformation
        fwg = hyperelastic.GeneralizedInvariantsFramework(tod, fun=fun, exponent=k)

        umat = hyperelastic.DeformationSpace(fwg)
        P = umat.gradient([defgrad, None])[0]

        force = P[0, 0] - P[2, 2] * 1 / (λ * np.sqrt(λ))
        dforce = np.diff(force)[0] / ϵ

        plt.plot(λ, force, label=f"k={k}")
        plt.ylim(0, 5)
        plt.legend()

        assert np.isclose(dforce, 3)


if __name__ == "__main__":
    test_generalized()



from ._experiment import Experiment
from ._load_case import Biaxial, IncompressibleHomogeneousStretch, Planar, Uniaxial
from ._optimize import Optimize
from ._simulation import Simulation

__all__ = [
    "Experiment",
    "Uniaxial",
    "Biaxial",
    "Planar",
    "IncompressibleHomogeneousStretch",
    "Optimize",
    "Simulation",
]
